import codecs
import json
import tarfile
from logging import getLogger

from qtpy import QtWidgets, QtCore
from bw2io import backup

from activity_browser import application
from activity_browser.mod import bw2data as bd
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons
from activity_browser.ui.threading import ABThread

log = getLogger(__name__)


class ProjectImport(ABAction):
    """
    ABAction to import a new project. Prompts the user to select a file. Imports the project name from the file as a
    suggestion. Prompts user to either accept the name or change it. If the name already exists, try again. Else,
    perform the import in a separate thread and show a progress dialog until it is finished. Finally, move to the newly
    imported project.
    """
    icon = qicons.import_db
    text = "&Import a project..."
    tool_tip = "Import project from a file"

    @classmethod
    @exception_dialogs
    def run(cls):
        """Import a project into AB based on file chosen by user."""

        # get the path from the user
        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            parent=application.main_window,
            caption='Choose project file to import',
            filter='Tar archive (*.tar.gz);; All files (*.*)'
        )
        if not path: return

        # create a name suggestion based on the file name
        suggestion = cls.get_project_name(path)

        # get a new project name from the user:
        while True:
            project_name, _ = QtWidgets.QInputDialog.getText(
                application.main_window,
                'Choose project name',
                'Choose a name for your project',
                text=suggestion
            )

            if not project_name: return

            if project_name in bd.projects:
                # this name already exists, inform user and ask again.
                QtWidgets.QMessageBox.information(
                    application.main_window,
                    "Not possible.",
                    "A project with this name already exists."
                )
            else: break

        # setup dialog
        progress = QtWidgets.QProgressDialog(
            parent=application.main_window,
            labelText="Importing project",
            maximum=0
        )
        progress.setCancelButton(None)
        progress.setWindowTitle("Importing project")
        progress.setWindowFlag(QtCore.Qt.WindowContextHelpButtonHint, False)
        progress.setWindowFlag(QtCore.Qt.WindowCloseButtonHint, False)
        progress.findChild(QtWidgets.QProgressBar).setTextVisible(False)
        progress.resize(400, 100)
        progress.show()

        # setup the import
        thread = ImportThread(application)
        setattr(thread, "path", path)
        setattr(thread, "project_name", project_name)

        thread.finished.connect(lambda: progress.deleteLater())
        thread.finished.connect(lambda: bd.projects.set_current(project_name))

        # start the import
        thread.start()

    @staticmethod
    def get_project_name(fp):
        reader = codecs.getreader("utf-8")
        # See https://stackoverflow.com/questions/68997850/python-readlines-with-tar-file-gives-streamerror-seeking-backwards-is-not-al/68998071#68998071
        with tarfile.open(fp, "r:gz") as tar:
            for member in tar:
                if member.name[-17:] == "project-name.json":
                    return json.load(reader(tar.extractfile(member)))["name"]
            raise ValueError("Couldn't find project name file in archive")


class ImportThread(ABThread):

    def run_safely(self):
        log.debug('Starting project import:'
                  f'\nPATH: {self.path}'
                  f'\nNAME: {self.project_name}')
        backup.restore_project_directory(fp=self.path, project_name=self.project_name)
        log.info(f"Project `{self.project_name}` imported.")




from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons

from .project_switch import ProjectSwitch


class ProjectDuplicate(ABAction):
    """
    ABAction to duplicate a project. Asks the user for a new name. Returns if no name is given, the user cancels, or
    when the name is already in use by another project. Else, instructs the ProjectController to duplicate the current
    project to the new name.
    """

    icon = qicons.copy
    text = "Duplicate"
    tool_tip = "Duplicate the project"

    @staticmethod
    @exception_dialogs
    def run(name: str = None):
        if name is None:
            name = bd.projects.current

        new_name, ok = QtWidgets.QInputDialog.getText(
            application.main_window,
            "Duplicate current project",
            f"Duplicate project ({name}) to new name:"
            + " " * 10,
        )

        if not ok or not new_name:
            return

        if new_name in bd.projects:
            QtWidgets.QMessageBox.information(
                application.main_window,
                "Not possible.",
                "A project with this name already exists.",
            )
            return

        if name != bd.projects.current:
            bd.projects.set_current(name, update=False)
        bd.projects.copy_project(new_name, switch=False)  # don't switch because it will auto-update bw2 projects
        ProjectSwitch.run(new_name)  # switch using the action instead



from typing import List
from logging import getLogger

from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons

log = getLogger(__name__)


class MethodDelete(ABAction):
    """
    ABAction to remove one or multiple methods. First check whether the method is a node or leaf. If it's a node, also
    include all underlying methods. Ask the user for confirmation, and return if canceled. Otherwise, remove all found
    methods.
    """

    icon = qicons.delete
    text = "Delete Impact Category"

    @staticmethod
    @exception_dialogs
    def run(methods: List[tuple]):
        # check whether we're dealing with a leaf or node. If it's a node, select all underlying methods for deletion
        all_methods = [bd.Method(method) for method in methods]

        if len(all_methods) == 1:
            warning_text = f"Are you sure you want to delete this method?\n\n{methods[0]}"
        else:
            warning_text = f"Are you sure you want to delete {len(all_methods)} methods?"

        # warn the user about the pending deletion
        warning = QtWidgets.QMessageBox.warning(
            application.main_window,
            "Deleting Method",
            warning_text,
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No,
        )
        # return if the users cancels
        if warning == QtWidgets.QMessageBox.No:
            return

        # instruct the controller to delete the selected methods
        for method in all_methods:
            method.deregister()
            log.info(f"Deleted method {method.name}")



from logging import getLogger
from time import time

from qtpy.QtCore import QObject, Signal, SignalInstance
from blinker import signal as blinker_signal

log = getLogger(__name__)


class NodeSignals(QObject):
    changed: SignalInstance = Signal(object, object)
    deleted: SignalInstance = Signal(object)
    database_change: SignalInstance = Signal(object, object)
    code_change: SignalInstance = Signal(object, object)


class EdgeSignals(QObject):
    changed: SignalInstance = Signal(object, object)
    deleted: SignalInstance = Signal(object)
    recalculated: SignalInstance = Signal()


class MethodSignals(QObject):
    changed: SignalInstance = Signal(object)
    deleted: SignalInstance = Signal(object)


class ParameterSignals(QObject):
    changed: SignalInstance = Signal(object, object)
    deleted: SignalInstance = Signal(object)
    recalculated: SignalInstance = Signal()


class DatabaseSignals(QObject):
    written: SignalInstance = Signal(object)
    reset: SignalInstance = Signal(object)
    deleted: SignalInstance = Signal(str)


class ProjectSignals(QObject):
    changed: SignalInstance = Signal(object, object)  # Project changed | new project dataset, old project dataset
    created: SignalInstance = Signal()
    deleted: SignalInstance = Signal(str)


class MetaSignals(QObject):
    databases_changed: SignalInstance = Signal(object, object)
    methods_changed: SignalInstance = Signal(object, object)
    calculation_setups_changed: SignalInstance = Signal(object, object)


class ABSignals(QObject):
    """Signals used for the Activity Browser should be defined here.
    While arguments can be passed to signals, it is good practice not to do this if possible.
    Every signal should have a comment (no matter how descriptive the name of the signal) that describes what a
    signal is used for and after a pipe (|), what variables are sent, if any.
    """
    node = NodeSignals()
    edge = EdgeSignals()
    method = MethodSignals()
    database = DatabaseSignals()
    project = ProjectSignals()
    meta = MetaSignals()
    parameter = ParameterSignals()

    import_project = Signal()  # Import a project
    export_project = Signal()  # Export the current project
    database_selected = Signal(str)  # This database was selected (opened) | name of database
    database_read_only_changed = Signal(str, bool)  # The read_only state of database changed | name of database, read-only state
    database_tab_open = Signal(str)  # This database tab is being viewed by user | name of database
    add_activity_to_history = Signal(tuple)
    safe_open_activity_tab = Signal(tuple)  # Open activity details tab in read-only mode | key of activity
    unsafe_open_activity_tab = Signal(tuple)  # Open activity details tab in editable mode | key of activity
    close_activity_tab = Signal(tuple)  # Close this activity details tab | key of activity
    open_activity_graph_tab = Signal(tuple)  # Open the graph-view tab | key of activity
    edit_activity = Signal(str)  # An activity in this database may now be edited | name of database
    added_parameter = Signal(str, str, str)  # This parameter has been added | name of the parameter, amount, type (project, database or activity)
    parameters_changed = Signal()  # The parameters have changed
    parameter_scenario_sync = Signal(int, object, bool)  # Synchronize this data for table | index of the table, dataframe with scenario data, include default scenario
    parameter_superstructure_built = Signal(int, object)  # Superstructure built from scenarios | index of the table, dataframe with scenario data
    set_default_calculation_setup = Signal()  # Show the default (first) calculation setup
    calculation_setup_changed = Signal()  # Calculation setup was changed
    calculation_setup_selected = Signal(str)  # This calculation setup was selected (opened) | name of calculation setup
    lca_calculation = Signal(dict)  # Generate a calculation setup | dictionary with name, type (simple/scenario) and potentially scenario data
    delete_method = Signal(tuple, str)  # Delete this method | tuple of impact category, level of tree OR the proxy
    method_selected = Signal(tuple)  # This method was selected (opened) | tuple of method
    monte_carlo_finished = Signal()  # The monte carlo calculations are finished
    new_statusbar_message = Signal(str)  # Update the statusbar this message | message
    restore_cursor = Signal()  # Restore the cursor to normal
    project_updates_available = Signal(str, int)  # Project name and number of updates available
    toggle_show_or_hide_tab = Signal(str)  # Show/Hide the tab with this name | name of tab
    show_tab = Signal(str)  # Show this tab | name of tab
    hide_tab = Signal(str)  # Hide this tab | name of tab
    hide_when_empty = Signal()  # Show/Hide tab when it has/does not have sub-tabs
    plugin_selected = Signal(str, bool)  # This plugin was/was not selected | name of plugin, selected state

    def __getattribute__(self, item):
        """Delayed loading of connecting to the brighway signals"""
        setattr(ABSignals, "__getattribute__", super().__getattribute__)
        import bw2data as bd

        self._project_dataset = bd.projects.dataset

        self._connect_bw_signals()
        return super().__getattribute__(item)

    def _connect_bw_signals(self):
        from bw2data import signals, Method
        from bw2data.meta import databases, methods, calculation_setups

        patch_methods_datastore()
        patch_projects()

        signals.signaleddataset_on_save.connect(self._on_signaleddataset_on_save)
        signals.signaleddataset_on_delete.connect(self._on_signaleddataset_on_delete)
        signals.on_activity_database_change.connect(self._on_activity_database_change)
        signals.on_activity_code_change.connect(self._on_activity_code_change)

        signals.on_database_delete.connect(self._on_database_delete)
        signals.on_database_reset.connect(self._on_database_reset)
        signals.on_database_write.connect(self._on_database_write)

        signals.project_changed.connect(self._on_project_changed)
        signals.project_created.connect(self._on_project_created)

        signals.on_activity_parameter_recalculate.connect(self._on_parameter_recalculate)
        signals.on_database_parameter_recalculate.connect(self._on_parameter_recalculate)
        signals.on_project_parameter_recalculate.connect(self._on_parameter_recalculate)
        signals.on_activity_parameter_recalculate_exchanges.connect(self._on_parameterized_exchange_recalculate)

        databases._save_signal.connect(self._on_database_metadata_change)
        setattr(methods, "_save_signal", blinker_signal("ab.patched_methods"))
        methods._save_signal.connect(self._on_methods_metadata_change)
        setattr(calculation_setups, "_save_signal", blinker_signal("ab.patched_calculation_setups"))
        calculation_setups._save_signal.connect(self._on_cs_metadata_change)

        Method._write_signal.connect(self._on_method_write)
        Method._deregister_signal.connect(self._on_method_deregister)

    def _on_signaleddataset_on_save(self, sender, old, new):
        from bw2data.backends import ActivityDataset, ExchangeDataset
        from bw2data.parameters import ProjectParameter, DatabaseParameter, ActivityParameter

        if isinstance(new, ActivityDataset):
            t = time()
            self.node.changed.emit(new, old)
            log.debug(f"Activity changed signal completed in {time() - t:.2f} seconds")
        elif isinstance(new, ExchangeDataset):
            t = time()
            self.edge.changed.emit(new, old)
            log.debug(f"Exchange changed signal completed in {time() - t:.2f} seconds")
        elif isinstance(new, (ProjectParameter, DatabaseParameter, ActivityParameter)):
            t = time()
            self.parameter.changed.emit(new, old)
            log.debug(f"Parameter changed signal completed in {time() - t:.2f} seconds")
        else:
            log.debug(f"Unknown dataset type changed: {type(new)}")

    def _on_signaleddataset_on_delete(self, sender, old):
        from bw2data.backends import ActivityDataset, ExchangeDataset
        from bw2data.parameters import ProjectParameter, DatabaseParameter, ActivityParameter

        if isinstance(old, ActivityDataset):
            t = time()
            self.node.deleted.emit(old)
            log.debug(f"Activity deleted signal completed in {time() - t:.2f} seconds")
        elif isinstance(old, ExchangeDataset):
            t = time()
            self.edge.deleted.emit(old)
            log.debug(f"Exchange deleted signal completed in {time() - t:.2f} seconds")
        elif isinstance(old, (ProjectParameter, DatabaseParameter, ActivityParameter)):
            t = time()
            self.parameter.deleted.emit(old)
            log.debug(f"Parameter deleted signal completed in {time() - t:.2f} seconds")
        else:
            log.debug(f"Unknown dataset type deleted: {type(old)}")

    def _on_activity_database_change(self, sender, old, new):
        t = time()
        self.node.database_change.emit(old, new)
        log.debug(f"Activity db changed signal completed in {time() - t:.2f} seconds")

    def _on_activity_code_change(self, sender, old, new):
        t = time()
        self.node.code_change.emit(old, new)
        log.debug(f"Activity code changed signal completed in {time() - t:.2f} seconds")

    def _on_database_delete(self, sender, name):
        t = time()
        self.database.deleted.emit(name)
        log.debug(f"Database deleted signal completed in {time() - t:.2f} seconds")

    def _on_database_reset(self, sender, name):
        from bw2data import Database
        t = time()
        self.database.reset.emit(Database(name))
        log.debug(f"Database reset signal completed in {time() - t:.2f} seconds")

    def _on_database_write(self, sender, name):
        from bw2data import Database
        t = time()
        self.database.written.emit(Database(name))
        log.debug(f"Database write signal completed in {time() - t:.2f} seconds")

    def _on_project_changed(self, ds):
        t = time()
        self.project.changed.emit(ds, self._project_dataset)
        self._project_dataset = ds
        log.debug(f"Project changed signal completed in {time() - t:.2f} seconds")

    def _on_project_created(self, ds):
        t = time()
        self.project.created.emit()
        log.debug(f"Project created signal completed in {time() - t:.2f} seconds")

    def _on_database_metadata_change(self, sender, old, new):
        t = time()
        self.meta.databases_changed.emit(old, new)
        log.debug(f"DB metadata changed signal completed in {time() - t:.2f} seconds")

    def _on_methods_metadata_change(self, sender, old, new):
        t = time()
        self.meta.methods_changed.emit(old, new)
        log.debug(f"Methods metadata changed signal completed in {time() - t:.2f} seconds")

    def _on_cs_metadata_change(self, sender, old, new):
        t = time()
        self.meta.calculation_setups_changed.emit(old, new)
        log.debug(f"CS metadata changed signal completed in {time() - t:.2f} seconds")

    def _on_method_write(self, sender):
        t = time()
        self.method.changed.emit(sender)
        log.debug(f"Method changed signal completed in {time() - t:.2f} seconds")

    def _on_method_deregister(self, sender):
        t = time()
        self.method.deleted.emit(sender)
        log.debug(f"Method deleted signal completed in {time() - t:.2f} seconds")

    def _on_parameter_recalculate(self, sender, *args, **kwargs):
        t = time()
        self.parameter.recalculated.emit()
        log.debug(f"Param recalculated signal completed in {time() - t:.2f} seconds")

    def _on_parameterized_exchange_recalculate(self, sender, *args, **kwargs):
        t = time()
        self.edge.recalculated.emit()
        log.debug(f"Param exchange recalculated signal completed in {time() - t:.2f} seconds")


def patch_methods_datastore():
    from bw2data import Method

    def write(self, data, process=True):
        original_write(self, data, process)
        self._write_signal.send(self)

    def deregister(self):
        original_deregister(self)
        self._deregister_signal.send(self)

    original_write = Method.write
    original_deregister = Method.deregister

    setattr(Method, "write", write)
    setattr(Method, "deregister", deregister)

    setattr(Method, "_write_signal", blinker_signal("ab.patched_method_write"))
    setattr(Method, "_deregister_signal", blinker_signal("ab.patched_method_deregister"))


def patch_projects():
    from bw2data.project import ProjectManager

    def delete_project(self, name=None, delete_dir=False):
        original_delete(self, name, delete_dir)
        signals.project.deleted.emit(name)

    original_delete = ProjectManager.delete_project

    setattr(ProjectManager, "delete_project", delete_project)


signals = ABSignals()



import json
import time
from typing import List, Optional
from logging import getLogger

import bw2calc as bc
import bw2data as bd
from qtpy import QtWidgets
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QComboBox
from bw_graph_tools.graph_traversal import (
    SameNodeEachVisitGraphTraversal,
    SameNodeEachVisitTaggedGraphTraversal,
    GraphTraversalSettings,
    TaggedGraphTraversalSettings,
)
from bw_graph_tools.graph_traversal.graph_objects import (
    Node as GraphNode,
    Edge as GraphEdge,
    GroupedNodes as GraphGroupedNodes,
)

from activity_browser import signals
from activity_browser.mod import bw2data as bd
from bw2data.backends import ActivityDataset
from activity_browser.utils import get_base_path
from .base import BaseGraph, BaseNavigatorWidget
from ..widgets.combobox import CheckableComboBox
from ...bwutils import AB_metadata
from ...bwutils.commontasks import identify_activity_type

log = getLogger(__name__)


class TreeNavigatorWidget(BaseNavigatorWidget):
    HELP_TEXT = """
    LCA Dynamic Tree Navigator:

    Red flows: Impacts
    Green flows: Avoided impacts

    """
    HTML_FILE = str(get_base_path().joinpath("static", "tree_navigator.html").resolve())

    def __init__(self, cs_name, parent=None):
        super().__init__(parent, css_file="tree_navigator.css")

        self.cache = {}  # we cache the calculated data to improve responsiveness
        self.parent = parent
        self.has_scenarios = self.parent.has_scenarios
        self.cs = cs_name
        self.selected_db = None
        self.has_rendered_once = False
        self.func_units = []
        self.methods = []
        self.scenarios = []
        self.graph = Graph()

        # Additional Qt objects
        self.scenario_label = QtWidgets.QLabel("Scenario: ")
        self.func_unit_cb = QtWidgets.QComboBox()
        self.method_cb = QtWidgets.QComboBox()
        self.scenario_cb = QtWidgets.QComboBox()
        self.tag_cb = CheckableComboBox()
        self.cutoff_sb = QtWidgets.QDoubleSpinBox()
        self.max_calc_sb = QtWidgets.QDoubleSpinBox()
        self.button_calculate = QtWidgets.QPushButton("Calculate")
        self.layout = QtWidgets.QVBoxLayout()

        # graph
        self.draw_graph()
        self.construct_layout()
        self.connect_signals()

    @Slot(name="loadFinishedHandler")
    def load_finished_handler(self) -> None:
        self.send_json()

    def connect_signals(self):
        super().connect_signals()
        self.button_calculate.clicked.connect(self.new_tree)
        signals.database_selected.connect(self.set_database)
        # checkboxes
        self.func_unit_cb.currentIndexChanged.connect(self.new_tree)
        self.method_cb.currentIndexChanged.connect(self.new_tree)
        self.scenario_cb.currentIndexChanged.connect(self.new_tree)
        self.tag_cb.onHidePopup.connect(self.new_tree)
        self.bridge.update_graph.connect(self.update_graph)

    def construct_layout(self) -> None:
        """Layout of Sankey Navigator"""
        super().construct_layout()
        self.label_help.setVisible(False)

        # Layout Reference Flows and Impact Categories
        grid_lay = QtWidgets.QGridLayout()
        grid_lay.addWidget(QtWidgets.QLabel("Reference flow: "), 0, 0)

        grid_lay.addWidget(self.scenario_label, 1, 0)
        grid_lay.addWidget(QtWidgets.QLabel("Impact indicator: "), 2, 0)
        grid_lay.addWidget(QtWidgets.QLabel("Tag System: "), 2, 2)

        self.update_calculation_setup()

        grid_lay.addWidget(self.func_unit_cb, 0, 1, 1, 3)
        grid_lay.addWidget(self.scenario_cb, 1, 1)
        grid_lay.addWidget(self.method_cb, 2, 1)
        grid_lay.addWidget(self.tag_cb, 2, 3)

        # cut-off
        grid_lay.addWidget(QtWidgets.QLabel("Cutoff: "), 2, 4)
        self.cutoff_sb.setRange(0.0, 1.0)
        self.cutoff_sb.setSingleStep(0.01)
        self.cutoff_sb.setDecimals(3)
        self.cutoff_sb.setValue(0.05)
        self.cutoff_sb.setKeyboardTracking(False)
        grid_lay.addWidget(self.cutoff_sb, 2, 5)

        # max-iterations of graph traversal
        grid_lay.addWidget(QtWidgets.QLabel("Calculation depth: "), 2, 6)
        self.max_calc_sb.setRange(1, 2000)
        self.max_calc_sb.setSingleStep(50)
        self.max_calc_sb.setDecimals(0)
        self.max_calc_sb.setValue(250)
        self.max_calc_sb.setKeyboardTracking(False)
        grid_lay.addWidget(self.max_calc_sb, 2, 7)

        grid_lay.setColumnStretch(6, 1)
        hlay = QtWidgets.QHBoxLayout()
        hlay.addLayout(grid_lay)

        # Controls Layout
        # hl_controls = QtWidgets.QHBoxLayout()
        grid_lay.addWidget(self.button_calculate, 0, 5)
        grid_lay.addWidget(self.button_refresh, 0, 6)
        grid_lay.addWidget(self.button_toggle_help, 0, 7)
        # hl_controls.addStretch(1)

        # Layout
        self.layout.addLayout(hlay)
        # self.layout.addLayout(hl_controls)
        self.layout.addWidget(self.label_help)
        self.layout.addWidget(self.view)
        self.setLayout(self.layout)

    def get_scenario_labels(self) -> List[str]:
        """Get scenario labels if scenario is used."""
        return self.parent.mlca.scenario_names if self.has_scenarios else []

    def configure_scenario(self):
        """Determine if scenario Qt widgets are visible or not and retrieve
        scenario labels for the selection drop-down box.
        """
        self.scenario_cb.setVisible(self.has_scenarios)
        self.scenario_label.setVisible(self.has_scenarios)
        if self.has_scenarios:
            self.scenarios = self.get_scenario_labels()
            self.update_combobox(self.scenario_cb, self.scenarios)

    @staticmethod
    def update_combobox(box: QComboBox, labels: List[str]) -> None:
        """Update the combobox menu."""
        box.blockSignals(True)
        box.clear()
        box.insertItems(0, labels)
        box.blockSignals(False)

    def update_calculation_setup(self, cs_name=None) -> None:
        """Update Calculation Setup, reference flows and impact categories, and dropdown menus."""
        # block signals
        block_signals = [self.func_unit_cb, self.method_cb, self.tag_cb]
        for b in block_signals:
            b.blockSignals(True)

        self.cs = cs_name or self.cs
        self.func_units = [
            {bd.get_activity(k): v for k, v in fu.items()}
            for fu in bd.calculation_setups[self.cs]["inv"]
        ]
        self.methods = bd.calculation_setups[self.cs]["ia"]
        self.func_unit_cb.clear()
        fu_acts = [list(fu.keys())[0] for fu in self.func_units]
        self.func_unit_cb.addItems(
            [f"{repr(a)} | {a._data.get('database')}" for a in fu_acts]
        )
        self.configure_scenario()
        self.method_cb.clear()
        self.method_cb.addItems([repr(m) for m in self.methods])

        # tags
        self.tag_cb.clear()
        self.tag_cb.addItems(sorted(list(AB_metadata.get_tag_names())))

        # unblock signals
        for b in block_signals:
            b.blockSignals(False)

    def new_tree(self) -> None:
        """(re)-generate the tree diagram."""
        demand_index = self.func_unit_cb.currentIndex()
        method_index = self.method_cb.currentIndex()
        self.update_tree(
            self.func_units[demand_index],
            self.methods[method_index],
            demand_index=demand_index,
            method_index=method_index,
            scenario_index=self.scenario_cb.currentIndex() if self.has_scenarios else None,
            scenario_lca=bool(self.has_scenarios),
            cut_off=self.cutoff_sb.value(),
            max_calc=int(self.max_calc_sb.value()),
            tags=self.tag_cb.currentData(),
        )

    def update_tree(
        self,
        demand: dict,
        method: tuple,
        demand_index: int = None,
        method_index: int = None,
        scenario_index: int = None,
        scenario_lca: bool = False,
        cut_off=0.05,
        max_calc=100,
        tags=None,
    ) -> None:
        """Calculate LCA, do graph traversal, get JSON graph data for this, and send to javascript."""

        # the cache key consists of demand/method/scenario indices (index of item in the relevant tables),
        # the cutoff, max_calc.
        # together, these are unique.
        cache_key = (
            demand_index,
            method_index,
            scenario_index,
            cut_off,
            max_calc,
            str(tags),
        )
        if data := self.cache.get(cache_key, False):
            # this Graph is already cached, generate the tree with Graph cached data
            log.debug(f"CACHED tree for: {demand}, {method}, key: {cache_key}")
            self.graph.new_graph(data)
            self.has_rendered_once = bool(self.graph.json_data)
            self.send_json()
            return

        start = time.time()
        log.debug(f"CALCULATE tree for: {demand}, {method}, key: {cache_key}")

        try:
            if scenario_lca:
                self.parent.mlca.update_lca_calculation_for_sankey(
                    scenario_index, demand, method_index
                )

            if not hasattr(self, "cached_lca"):
                fu, data_objs, _ = bd.prepare_lca_inputs(demand=demand, method=method)
                self.cached_lca = bc.LCA(demand=fu, data_objs=data_objs)
                self.cached_lca.lci(factorize=True)
                self.cached_lca.lcia()
            if tags:
                data = SameNodeEachVisitTaggedGraphTraversal(
                    lca=self.cached_lca,
                    settings=TaggedGraphTraversalSettings(
                        tags=tags, cutoff=cut_off, max_calc=max_calc
                    ),
                )
            else:
                data = SameNodeEachVisitGraphTraversal(
                    lca=self.cached_lca,
                    settings=GraphTraversalSettings(
                        cutoff=cut_off, max_calc=max_calc
                    ),
                )
            data.traverse(depth=2)

            # store the metadata from this calculation
            data.metadata = {
                "unit": bd.methods[method]["unit"],
            }
        except (ValueError, ZeroDivisionError) as e:
            QtWidgets.QMessageBox.information(
                None, "Nonsensical numeric result.", str(e)
            )
        log.debug(f"Completed graph traversal ({round(time.time() - start, 2)} seconds")

        # cache the generated Graph data
        self.cache[cache_key] = data

        # generate the new Graph
        self.graph.new_graph(data)
        self.has_rendered_once = bool(self.graph.json_data)
        self.send_json()

    def set_database(self, name):
        """Saves the currently selected database for graphing a random activity"""
        self.selected_db = name

    def random_graph(self) -> None:
        """Show graph for a random activity in the currently loaded database."""
        if self.selected_db:
            method = bd.methods.random()
            act = bd.Database(self.selected_db).random()
            demand = {act: 1.0}
            self.update_tree(demand, method)
        else:
            QtWidgets.QMessageBox.information(
                None, "Not possible.", "Please load a database first."
            )

    @Slot(object, name="update_graph")
    def update_graph(self, click_dict: dict) -> None:
        """Update the graph with the specified JSON data."""
        traversed = self.graph.state_graph.traverse_from_node(click_dict["id"])
        if not traversed:
            # nothing has changed
            return
        self.graph.json_data = Graph.get_json_data(self.graph.state_graph)
        self.send_json()


class Graph(BaseGraph):
    """
    Python side representation of the graph.
    Functionality for graph navigation (e.g. adding and removing nodes).
    A JSON representation of the graph (edges and nodes) enables its use in javascript/html/css.
    """

    def __init__(self):
        super().__init__()
        self.state_graph: Optional["SameNodeEachVisitGraphTraversal"] = None

    @staticmethod
    def get_data_from_state_graph(state_graph: "SameNodeEachVisitGraphTraversal"):
        return {
            "nodes": state_graph.nodes,
            "edges": state_graph.edges,
            "flows": state_graph.flows,
            "calculation_count": state_graph.calculation_count.value,
            "metadata": state_graph.metadata,
        }

    def new_graph(self, state_graph: "SameNodeEachVisitGraphTraversal"):
        self.state_graph = state_graph
        self.json_data = Graph.get_json_data(state_graph)
        self.update()

    @staticmethod
    def get_json_data(state_graph: "SameNodeEachVisitGraphTraversal") -> str:
        """Transform graph traversal output to JSON data.

        We use the [dagre](https://github.com/dagrejs/dagre) javascript library for rendering directed graphs. We need to provide the following:

        ```python
        {
            'max_impact': float,  # Total LCA score,
            'title': str,  # Graph title
            'edges': [{
                'source_id': int,  # Unique ID of producer of material or energy in graph
                'target_id': int,  # Unique ID of consumer of material or energy in graph
                'weight': float,  #  In graph units, relative to `max_edge_width`
                'label': str,  # HTML label
                'product': str,  # The label of the flowing material or energy
                'class': str,  # "benefit" or "impact"; controls styling
                'label': str,  # HTML label
                'toottip': str,  # HTML tooltip
            }],
            'nodes': [{
                'direct_emissions_score_normalized': float,  # Fraction of total LCA score from direct emissions
                'product': str,  # Reference product label, if any
                'location': str,  # Location, if any
                'id': int,  # Graph traversal ID
                'database_id': int,  # Node ID in SQLite database
                'database': str,  # Database name
                'class': str,  # Enumerated set of class label strings
                'label': str,  # HTML label including name and location
                'toottip': str,  # HTML tooltip
            }],
        }

        ```

        """
        lca_score = state_graph.lca.score
        lcia_unit = state_graph.metadata["unit"]
        demand = state_graph.lca.demand

        def convert_edge_to_json(
            edge: GraphEdge,
            nodes: dict[int, GraphNode],
            total_score: float,
            lcia_unit: str,
            max_edge_width: int = 40,
        ) -> dict:
            cum_score = nodes[edge.producer_unique_id].cumulative_score
            node = nodes[edge.producer_unique_id]
            if isinstance(node, GraphGroupedNodes):
                unit = ""
            else:
                unit = bd.get_node(
                    id=nodes[edge.producer_unique_id].reference_product_datapackage_id
                ).get("unit", "(unknown)")
            return {
                "source_id": edge.producer_unique_id,
                "target_id": edge.consumer_unique_id,
                "amount": edge.amount,
                "weight": abs(cum_score / total_score) * max_edge_width,
                "label": f"{round(cum_score, 3)} {lcia_unit}",
                "class": "benefit" if cum_score < 0 else "impact",
                "tooltip": f"<b>{round(cum_score, 3)} {lcia_unit}</b> ({edge.amount:.2g} {unit})",
            }

        def convert_node_to_json(
            graph_node: GraphNode,
            total_score: float,
            fu: dict,
            lcia_unit: str,
            max_name_length: int = 20,
        ) -> dict:
            expanded = graph_node.unique_id in state_graph.visited_nodes
            if isinstance(graph_node, GraphGroupedNodes):
                data = {
                    "direct_emissions_score_normalized": graph_node.direct_emissions_score
                    / (total_score or 1),
                    "direct_emissions_score": graph_node.direct_emissions_score,
                    "cumulative_score": graph_node.cumulative_score,
                    "cumulative_score_normalized": graph_node.cumulative_score
                    / (total_score or 1),
                    "product": "",
                    "location": "",
                    "id": graph_node.unique_id,
                    "database_id": "",
                    "database": "",
                    "class": "",
                    "name": graph_node.label,
                    "expanded": expanded,
                }
            else:
                db_node = bd.get_node(id=graph_node.activity_datapackage_id)
                data = {
                    "direct_emissions_score_normalized": graph_node.direct_emissions_score
                    / (total_score or 1),
                    "direct_emissions_score": graph_node.direct_emissions_score,
                    "cumulative_score": graph_node.cumulative_score,
                    "cumulative_score_normalized": graph_node.cumulative_score
                    / (total_score or 1),
                    "product": db_node.get("reference product", ""),
                    "location": db_node.get("location", "(unknown)"),
                    "id": graph_node.unique_id,
                    "database_id": graph_node.activity_datapackage_id,
                    "database": db_node["database"],
                    "class": (
                        "demand"
                        if graph_node.activity_datapackage_id in fu
                        else identify_activity_type(db_node)
                    ),
                    "name": db_node.get("name", "(unnamed)"),
                    "expanded": expanded,
                }
            frac_dir_score = round(data["direct_emissions_score_normalized"] * 100, 2)
            dir_score = round(data["direct_emissions_score"], 3)
            frac_cum_score = round(data["cumulative_score_normalized"] * 100, 2)
            cum_score = round(data["cumulative_score"], 3)
            if isinstance(graph_node, GraphGroupedNodes):
                data["label"] = data["name"]
            else:
                data["label"] = "{}\n{}\n{}".format(
                    db_node["name"][:max_name_length], data["location"], frac_dir_score
                )
            data[
                "tooltip"
            ] = f"""
                <b>{data['name']}</b>
                <br>Individual impact: {dir_score} {lcia_unit} ({frac_dir_score}%)
                <br>Cumulative impact: {cum_score} {lcia_unit} ({frac_cum_score}%)
                <br>Expanded: {expanded}
            """
            return data

        json_data = {
            "nodes": [
                convert_node_to_json(node, lca_score, demand, lcia_unit)
                for idx, node in state_graph.nodes.items()
                if idx != -1
            ],
            "edges": [
                convert_edge_to_json(edge, state_graph.nodes, lca_score, lcia_unit)
                for edge in state_graph.edges
                if edge.producer_index != -1 and edge.consumer_index != -1
            ],
            "title": None,
        }

        return json.dumps(json_data)


def id_to_key(id):
    if isinstance(id, tuple):
        return id
    return ActivityDataset.get_by_id(id).database, ActivityDataset.get_by_id(id).code



from activity_browser.actions.base import ABAction, exception_dialogs

from activity_browser.ui.icons import qicons

from activity_browser.mod.bw2io.migrations import ab_create_core_migrations


class ToolsBW2IOCreateMigrations(ABAction):
    """
    ABAction to install default migrations from bw2io
    """

    icon = qicons.import_db
    text = "Install default bw2io migrations"

    @staticmethod
    @exception_dialogs
    def run():
        ab_create_core_migrations()



import itertools
import json
import os
from copy import deepcopy
from typing import Optional
from logging import getLogger

import networkx as nx
from qtpy import QtWidgets
from qtpy.QtCore import Slot

from activity_browser import signals
from bw2data import Database, get_activity, databases, Edge
from bw2data.backends import ExchangeDataset, ActivityDataset

from ...bwutils.commontasks import identify_activity_type, get_activity_name
from .base import BaseGraph, BaseNavigatorWidget

log = getLogger(__name__)


# TODO:
# save graph as image
# zoom reverse direction between canvas and minimap
# break long geographies into max length
# enable other layouts (e.g. force)
# random_graph should not work for biosphere
# a selection possibility method would be nice if many nodes are to be added up/downstream (now the only way is to open all and close those that one is not interested in)

# ISSUES:
# - tooltips show values, but these are not scaled to a product system, i.e. the do not make sense as a system


class GraphNavigatorWidget(BaseNavigatorWidget):
    HELP_TEXT = """
    How to use the Graph Navigator:

    EXPANSION MODE (DEFAULT):
    Click on activities to expand graph. 
    - click: expand upwards
    - click + shift: expand downstream
    - click + alt: delete activity

    Checkbox "Add only direct up-/downstream exchanges" - there are two ways to expand the graph: 
        1) adding direct up-/downstream nodes and connections (DEFAULT). 
        2) adding direct up-/downstream nodes and connections AS WELL as ALL OTHER connections between the activities in the graph. 
        The first option results in cleaner (but not complete) graphs.    

    Checkbox "Remove orphaned nodes": by default nodes that do not link to the central activity (see title) are removed (this may happen after deleting nodes). Uncheck to disable.

    Checkbox "Flip negative flows" (experimental): Arrows of negative product flows (e.g. from ecoinvent treatment activities or from substitution) can be flipped. 
    The resulting representation can be more intuitive for understanding the physical product flows (e.g. that wastes are outputs of activities and not negative inputs).   


    NAVIGATION MODE:
    Click on activities to jump to specific activities (instead of expanding the graph).
    """
    HTML_FILE = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "../../static/navigator.html"
    )

    def __init__(self, parent=None, key=None):
        super().__init__(parent, css_file="navigator.css")
        self.setObjectName(get_activity_name(get_activity(key), str_length=30))
        self.key = key
        self.tab = parent

        self.graph = Graph()

        # default settings
        self.navigation_label = itertools.cycle(
            ["Current mode: Expansion", "Current mode: Navigation"]
        )
        self.selected_db = None

        self.button_navigation_mode = QtWidgets.QPushButton(next(self.navigation_label))
        self.checkbox_direct_only = QtWidgets.QCheckBox(
            "Add only direct up-/downstream exchanges"
        )
        self.checkbox_remove_orphaned_nodes = QtWidgets.QCheckBox(
            "Remove orphaned nodes"
        )
        self.checkbox_flip_negative_edges = QtWidgets.QCheckBox("Flip negative flows")
        self.layout = QtWidgets.QVBoxLayout()

        # Prepare graph
        self.draw_graph()

        # Construct layout and set signals.
        self.construct_layout()
        self.update_graph_settings()
        self.connect_signals()

        if key:
            self.selected_db = key[0]
            self.new_graph(key)

    @Slot(name="loadFinishedHandler")
    def load_finished_handler(self) -> None:
        """Executed when webpage has been loaded for the first time or refreshed.
        This is needed to resend the json data the first time after the page has completely loaded.
        """
        # print(time.time(), ": load finished")
        self.send_json()

    def connect_signals(self):
        super().connect_signals()
        self.button_navigation_mode.clicked.connect(self.toggle_navigation_mode)
        # signals.database_selected.connect(self.set_database)
        self.bridge.update_graph.connect(self.update_graph)
        # checkboxes
        self.checkbox_direct_only.stateChanged.connect(self.update_graph_settings)
        self.checkbox_remove_orphaned_nodes.stateChanged.connect(
            self.update_graph_settings
        )
        self.checkbox_flip_negative_edges.stateChanged.connect(
            self.update_graph_settings
        )
        self.checkbox_flip_negative_edges.stateChanged.connect(self.reload_graph)
        databases.metadata_changed.connect(self.sync_graph)

    def sync_graph(self):
        """Sync the graph with the current project."""
        self.graph.update(delete_unstacked=False)
        self.send_json()
        try:
            self.setObjectName(get_activity_name(get_activity(self.key), str_length=30))
        except ActivityDataset.DoesNotExist:
            log.debug("Graph activity no longer exists. Closing tab.")
            self.tab.close_tab_by_tab_name(self.tab.get_tab_name(self))

    def construct_layout(self) -> None:
        """Layout of Graph Navigator"""
        self.label_help.setVisible(False)

        # checkbox all_exchanges_in_graph
        self.checkbox_direct_only.setChecked(True)
        self.checkbox_direct_only.setToolTip(
            "When adding activities, show product flows between ALL activities or just selected up-/downstream flows"
        )

        # checkbox remove orphaned nodes
        self.checkbox_remove_orphaned_nodes.setChecked(True)
        self.checkbox_remove_orphaned_nodes.setToolTip(
            "When removing activities, automatically remove those that have no further connection to the original product"
        )

        # checkbox flip negative edges
        self.checkbox_flip_negative_edges.setChecked(False)
        self.checkbox_flip_negative_edges.setToolTip(
            "Flip negative product flows (e.g. from ecoinvent treatment activities or from substitution)"
        )
        # Controls Layout
        hl_controls = QtWidgets.QHBoxLayout()
        hl_controls.addWidget(self.button_back)
        hl_controls.addWidget(self.button_forward)
        hl_controls.addWidget(self.button_navigation_mode)
        hl_controls.addWidget(self.button_refresh)
        hl_controls.addWidget(self.button_random_activity)
        hl_controls.addWidget(self.button_toggle_help)
        hl_controls.addStretch(1)

        # Checkboxes Layout
        hl_checkboxes = QtWidgets.QHBoxLayout()
        hl_checkboxes.addWidget(self.checkbox_direct_only)
        hl_checkboxes.addWidget(self.checkbox_remove_orphaned_nodes)
        hl_checkboxes.addWidget(self.checkbox_flip_negative_edges)
        hl_checkboxes.addStretch(1)

        # Layout
        self.layout.addLayout(hl_controls)
        self.layout.addLayout(hl_checkboxes)
        self.layout.addWidget(self.label_help)
        self.layout.addWidget(self.view)
        self.setLayout(self.layout)

    def update_graph_settings(self):
        self.graph.direct_only = self.checkbox_direct_only.isChecked()
        self.graph.remove_orphaned = self.checkbox_remove_orphaned_nodes.isChecked()
        self.graph.flip_negative_edges = self.checkbox_flip_negative_edges.isChecked()

    @property
    def is_expansion_mode(self) -> bool:
        return "Expansion" in self.button_navigation_mode.text()

    @Slot(name="toggleNavigationMode")
    def toggle_navigation_mode(self):
        mode = next(self.navigation_label)
        self.button_navigation_mode.setText(mode)
        log.info(f"Switched to: {mode}")
        self.checkbox_remove_orphaned_nodes.setVisible(self.is_expansion_mode)
        self.checkbox_direct_only.setVisible(self.is_expansion_mode)

    def new_graph(self, key: tuple) -> None:
        log.info(f"New Graph for key: {key}")
        self.graph.new_graph(key)
        self.send_json()

    @Slot(name="reload_graph")
    def reload_graph(self) -> None:
        signals.new_statusbar_message.emit("Reloading graph")
        self.graph.update(delete_unstacked=False)

    @Slot(object, name="update_graph")
    def update_graph(self, click_dict: dict) -> None:
        """
        Update graph based on user command (click+keyboard) and settings.
        Settings:
        - navigation or expansion mode
        - add all or only direct up/downstream nodes
        User commands:
        - mouse (left or right button)
        - additional keyboard keys (shift, alt)
        Behaviour: see HELP text
        """
        key = click_dict["key"]
        keyboard = click_dict["keyboard"]

        # interpret user command:
        if not self.is_expansion_mode:  # do not expand
            self.new_graph(key)
        else:
            if keyboard["alt"]:  # delete node
                log.info(f"Deleting node: {key}")
                self.graph.reduce_graph(key)
            else:  # expansion mode
                log.info(f"Expanding graph: {key}")
                if keyboard["shift"]:  # downstream expansion
                    log.info("Adding downstream nodes.")
                    self.graph.expand_graph(key, down=True)
                else:  # upstream expansion
                    log.info("Adding upstream nodes.")
                    self.graph.expand_graph(key, up=True)
            self.send_json()

    def set_database(self, name):
        """Saves the currently selected database for graphing a random activity"""
        self.selected_db = name

    @Slot(name="random_graph")
    def random_graph(self) -> None:
        """Show graph for a random activity in the currently loaded database."""
        if self.selected_db:
            self.new_graph(Database(self.selected_db).random().key)
        else:
            QtWidgets.QMessageBox.information(
                None, "Not possible.", "Please load a database first."
            )


class Graph(BaseGraph):
    """Python side representation of the graph.
    Functionality for graph navigation (e.g. adding and removing nodes).
    A JSON representation of the graph (edges and nodes) enables its use in javascript/html/css.
    """

    def __init__(self):
        super().__init__()
        self.central_activity = None
        self.nodes = None
        self.edges = None

        # some settings
        self.direct_only = True  # for a graph expansion: add only direct up-/downstream nodes instead of all connections between the activities in the graph
        self.remove_orphaned = True  # remove nodes that are isolated from the central_activity after a deletion
        self.flip_negative_edges = False  # show true flow direction of edges (e.g. for ecoinvent treatment activities, or substitutions)

    def update(self, delete_unstacked: bool = True) -> None:
        self.update_datasets()
        super().update(delete_unstacked)
        self.json_data = self.get_json_data()

    def update_datasets(self):
        """Update the activities in the graph."""
        try:
            self.nodes = [get_activity(act.key) for act in self.nodes]
            self.edges = [Edge(document=ExchangeDataset.get_by_id(exc._document.id)) for exc in self.edges]
        except (ActivityDataset.DoesNotExist, ExchangeDataset.DoesNotExist):
            try:
                get_activity(self.central_activity.key)  # test whether the activity still exists
                self.new_graph(self.central_activity.key)  # if so, create a new graph
            except ActivityDataset.DoesNotExist:
                log.warning("Graph activity no longer exists.")
                self.nodes = []
                self.edges = []

    def store_previous(self) -> None:
        self.stack.append((deepcopy(self.nodes), deepcopy(self.edges)))

    def store_future(self) -> None:
        self.forward_stack.append(self.stack.pop())
        self.nodes, self.edges = self.stack.pop()

    def retrieve_future(self) -> None:
        self.nodes, self.edges = self.forward_stack.pop()

    @staticmethod
    def upstream_and_downstream_nodes(key: tuple) -> (list, list):
        """Returns the upstream and downstream activity objects for a key."""
        activity = get_activity(key)
        upstream_nodes = [ex.input for ex in activity.technosphere()]
        downstream_nodes = [ex.output for ex in activity.upstream()]
        return upstream_nodes, downstream_nodes

    @staticmethod
    def upstream_and_downstream_exchanges(key: tuple) -> (list, list):
        """Returns the upstream and downstream Exchange objects for a key.

        act.upstream refers to downstream exchanges; brightway is confused here)
        """
        activity = get_activity(key)
        return [ex for ex in activity.technosphere()], [
            ex for ex in activity.upstream()
        ]

    @staticmethod
    def inner_exchanges(nodes: list) -> list:
        """Returns all exchanges (Exchange objects) between a list of nodes."""
        node_keys = set(node.key for node in nodes)
        exchanges = itertools.chain(node.technosphere() for node in nodes)
        return [
            ex
            for ex in exchanges
            if all(k in node_keys for k in (ex["input"], ex["output"]))
        ]

    def remove_outside_exchanges(self) -> None:
        """
        Ensures that all exchanges are exclusively between nodes of the graph
        (i.e. removes exchanges to previously existing nodes).
        """
        self.edges = [
            e for e in self.edges if all(k in self.nodes for k in (e.input, e.output))
        ]

    def new_graph(self, key: tuple) -> None:
        """Creates a new JSON graph showing the up- and downstream activities for the activity key passed.
        Args:
            key (tuple): activity key
        Returns:
                JSON data as a string
        """
        self.central_activity = get_activity(key)

        # add nodes
        up_nodes, down_nodes = Graph.upstream_and_downstream_nodes(key)
        self.nodes = [self.central_activity] + up_nodes + down_nodes

        # add edges
        # self.edges = self.inner_exchanges(self.nodes)
        up_exs, down_exs = Graph.upstream_and_downstream_exchanges(key)
        self.edges = up_exs + down_exs
        self.update()

    def expand_graph(self, key: tuple, up=False, down=False) -> None:
        """
        Adds up-, downstream, or both nodes to graph.
        Different behaviour for "direct nodes only" or "all nodes (inner exchanges)" modes.
        """
        up_nodes, down_nodes = Graph.upstream_and_downstream_nodes(key)

        # Add Nodes
        if up and not down:
            self.nodes = list(set(self.nodes + up_nodes))
        elif down and not up:
            self.nodes = list(set(self.nodes + down_nodes))
        elif up and down:
            self.nodes = list(set(self.nodes + up_nodes + down_nodes))

        # Add Edges / Exchanges
        if self.direct_only:
            up_exs, down_exs = Graph.upstream_and_downstream_exchanges(key)
            if up and not down:
                self.edges += up_exs
            elif down and not up:
                self.edges += down_exs
            elif up and down:
                self.edges += up_exs + down_exs
        else:  # all
            self.edges = Graph.inner_exchanges(self.nodes)
        self.update()

    def reduce_graph(self, key: tuple) -> None:
        """
        Deletes nodes from graph.
        Different behaviour for "direct nodes only" or "all nodes (inner exchanges)" modes.
        Can lead to orphaned nodes, which can be removed or kept.
        """
        if key == self.central_activity.key:
            log.warning("Cannot remove central activity.")
            return
        act = get_activity(key)
        self.nodes.remove(act)
        if self.direct_only:
            self.remove_outside_exchanges()
        else:
            self.edges = Graph.inner_exchanges(self.nodes)

        if self.remove_orphaned:  # remove orphaned nodes
            self.remove_orphaned_nodes()

        self.update()

    def remove_orphaned_nodes(self) -> None:
        """
        Remove orphaned nodes from graph using the networkx.
        Orphaned nodes are defined as having no path to the central_activity.
        """

        def format_as_weighted_edges(exchanges, activity_objects=False):
            """Returns the exchanges as a list of weighted edges (from, to, weight) for networkx."""
            if activity_objects:
                return ((ex.input, ex.output, ex.amount) for ex in exchanges)
            else:  # keys
                return ((ex["input"], ex["output"], ex["amount"]) for ex in exchanges)

        # construct networkx graph
        G = nx.MultiGraph()
        for node in self.nodes:
            G.add_node(node.key)
        G.add_weighted_edges_from(format_as_weighted_edges(self.edges))

        # identify orphaned nodes
        # checks each node in current dataset whether it is connected to central node
        # adds node_id of orphaned nodes to list
        orphaned_node_ids = (
            node
            for node in G.nodes
            if not nx.has_path(G, node, self.central_activity.key)
        )

        count = 1
        for count, key in enumerate(orphaned_node_ids, 1):
            act = get_activity(key)
            self.nodes.remove(act)
        log.info(f"Removed ORPHANED nodes: {count}")

        # update edges again to remove those that link to nodes that have been deleted
        self.remove_outside_exchanges()

    def get_json_data(self) -> Optional[str]:
        """
        Make the JSON graph data from a list of nodes and edges.

        Args:
            nodes: a list of nodes (Activity objects)
            edges: a list of edges (Exchange objects)
        Returns:
            A JSON representation of this.
        """
        if not self.nodes:
            log.info("Graph has no nodes (activities).")
            return

        data = {
            "nodes": [Graph.build_json_node(act) for act in self.nodes],
            "edges": [
                Graph.build_json_edge(exc, self.flip_negative_edges)
                for exc in self.edges
            ],
            "title": self.central_activity.get("reference product"),
        }
        # print("JSON DATA (Nodes/Edges):", len(nodes), len(edges))
        # print(data)
        return json.dumps(data)

    @staticmethod
    def build_json_node(act) -> dict:
        """Take an activity and return a valid JSON document."""
        return {
            "database": act.key[0],
            "id": act.key[1],
            "product": act.get("reference product") or act.get("name"),
            "name": act.get("name"),
            "location": act.get("location"),
            "class": identify_activity_type(act),
        }

    @staticmethod
    def build_json_edge(exc, flip_negative: bool) -> dict:
        """Take an exchange object and return a valid JSON document.

        ``flip_negative`` will change the direction of the edge to represent
        the correct physical flow direction. However, this is experimental,
        and may not be reflected in the actual display of the product/flow.
        """
        product = exc.input
        reference = product.get("reference product") or product.get("name")
        amount = exc.get("amount")
        from_act, to_act = exc.input, exc.output
        if flip_negative and amount < 0:
            from_act, to_act = to_act, from_act
            amount = abs(amount)
        return {
            "source_id": from_act.key[1],
            "target_id": to_act.key[1],
            "amount": amount,
            "unit": exc.get("unit"),
            "product": reference,
            "tooltip": "<b>{:.3g} {} of {}<b>".format(
                amount, exc.get("unit", ""), reference
            ),
        }



from logging import getLogger

import pandas as pd

from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt

from .item_model import ABItemModel

log = getLogger(__name__)


class ABTreeView(QtWidgets.QTreeView):
    # fired when the filter is applied, fires False when an exception happens during querying
    filtered: QtCore.SignalInstance = QtCore.Signal(bool)

    defaultColumnDelegates = {}

    class HeaderMenu(QtWidgets.QMenu):
        def __init__(self, pos: QtCore.QPoint, view: "ABTreeView"):
            super().__init__(view)

            model = view.model()

            col_index = view.columnAt(pos.x())
            col_name = model.columns()[col_index]

            search_box = QtWidgets.QLineEdit(self)
            search_box.setText(view.columnFilters.get(col_name, ""))
            search_box.setPlaceholderText("Search")
            search_box.selectAll()
            search_box.textChanged.connect(lambda query: view.setColumnFilter(col_name, query))
            widget_action = QtWidgets.QWidgetAction(self)
            widget_action.setDefaultWidget(search_box)
            self.addAction(widget_action)

            self.addAction(QtGui.QIcon(), "Group by column", lambda: model.group(col_index))
            self.addAction(QtGui.QIcon(), "Ungroup", model.ungroup)
            self.addAction(QtGui.QIcon(), "Clear column filter", lambda: view.setColumnFilter(col_name, ""))
            self.addAction(QtGui.QIcon(), "Clear all filters",
                lambda: [view.setColumnFilter(name, "") for name in list(view.columnFilters.keys())],
            )
            self.addSeparator()

            def toggle_slot(action: QtWidgets.QAction):
                index = action.data()
                hidden = view.isColumnHidden(index)
                view.setColumnHidden(index, not hidden)

            view_menu = QtWidgets.QMenu(view)
            view_menu.setTitle("View")
            self.view_actions = []

            for i in range(1, len(model.columns())):
                action = QtWidgets.QAction(model.columns()[i])
                action.setCheckable(True)
                action.setChecked(not view.isColumnHidden(i))
                action.setData(i)
                view_menu.addAction(action)
                self.view_actions.append(action)

            view_menu.triggered.connect(toggle_slot)

            self.addMenu(view_menu)

            search_box.setFocus()

    class ContextMenu(QtWidgets.QMenu):
        def __init__(self, pos, view):
            super().__init__(view)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setUniformRowHeights(True)

        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.showContextMenu)

        self.setSelectionBehavior(QtWidgets.QTreeView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QtWidgets.QTreeView.SelectionMode.ExtendedSelection)

        header = self.header()
        header.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        header.customContextMenuRequested.connect(self.showHeaderMenu)

        self.expanded_paths = set()
        self.expanded.connect(lambda index: self.expanded_paths.add(tuple(index.internalPointer().path())))
        self.collapsed.connect(lambda index: self.expanded_paths.discard(tuple(index.internalPointer().path())))

        self.columnFilters: dict[str, str] = {}  # dict[column_name, query] for filtering the dataframe
        self.allFilter: str = ""  # filter applied to the entire dataframe

    def setModel(self, model):
        if not isinstance(model, ABItemModel):
            raise TypeError("Model must be an instance of ABItemModel")
        super().setModel(model)

        model.modelReset.connect(self.expand_after_reset)
        model.modelAboutToBeReset.connect(self.clearColumnDelegates)
        model.modelReset.connect(self.setDefaultColumnDelegates)

        self.setDefaultColumnDelegates()

    def model(self) -> ABItemModel:
        return super().model()

    # === Functionality related to contextmenus

    def showContextMenu(self, pos):
        self.ContextMenu(pos, self).exec_(self.mapToGlobal(pos))

    def showHeaderMenu(self, pos):
        self.HeaderMenu(pos, self).exec_(self.mapToGlobal(pos))

    def setColumnFilter(self, column_name: str, query: str):
        """
        Set a filter for a specific column using a string query. If the query is empty remove the filter from the column
        """
        col_index = self.model().columns().index(column_name)

        if query:
            self.columnFilters[column_name] = query
            self.model().filtered_columns.add(col_index)
        elif column_name in self.columnFilters:
            del self.columnFilters[column_name]
            self.model().filtered_columns.discard(col_index)

        self.applyFilter()

    # === Functionality related to filtering

    def setAllFilter(self, query: str):
        self.allFilter = query
        self.applyFilter()

    def buildQuery(self) -> str:
        queries = ["(index == index)"]

        # query for the column filters
        for col in list(self.columnFilters):
            if col not in self.model().columns():
                del self.columnFilters[col]

        for col, query in self.columnFilters.items():
            q = f"({col}.astype('str').str.contains('{self.format_query(query)}'))"
            queries.append(q)

        # query for the all filter
        if self.allFilter.startswith('='):
            queries.append(f"({self.allFilter[1:]})")
        else:
            all_queries = []
            formatted_filter = self.format_query(self.allFilter)

            for i, col in enumerate(self.model().columns()):
                if self.isColumnHidden(i):
                    continue
                all_queries.append(f"(`{col}`.astype('str').str.contains('{formatted_filter}', False))")

            q = f"({' | '.join(all_queries)})"
            queries.append(q)

        query = " & ".join(queries)
        log.debug(f"{self.__class__.__name__} built query: {query}")

        return query

    def applyFilter(self):
        query = self.buildQuery()
        try:
            self.model().setQuery(query)
            self.filtered.emit(True)
        except Exception as e:
            log.info(f"{self.__class__.__name__} {type(e).__name__} in query: {e}")
            self.filtered.emit(False)

    @staticmethod
    def format_query(query: str) -> str:
        return query.translate(str.maketrans({'(': '\\(', ')': '\\)', "'": "\\'"}))

    # === Functionality related to setting the column delegates
    def clearColumnDelegates(self):
        for i in range(self.model().columnCount()):
            self.setItemDelegateForColumn(i, None)

    def setDefaultColumnDelegates(self):
        columns = self.model().columns()
        for i, col_name in enumerate(columns):
            if col_name in self.defaultColumnDelegates:
                delegate = self.defaultColumnDelegates[col_name](self)
                self.setItemDelegateForColumn(i, delegate)
            # elif col_name.startswith("property_"):
            #     self.setItemDelegateForColumn(i, self.propertyDelegate)

    # === Functionality related to saving and restoring the View's state

    def saveState(self) -> dict:
        if not self.model():
            return {}

        cols = self.model().columns()

        return {
            "columns": cols,
            "grouped_columns": [cols[i] for i in self.model().grouped_columns],
            "visible_columns": [cols[i] for i in range(len(cols)) if not self.isColumnHidden(i)],

            "expanded_paths": list(self.expanded_paths),

            "filters": self.columnFilters,
            "sort_column": cols[self.model().sort_column],
            "sort_ascending": self.model().sort_order == Qt.SortOrder.AscendingOrder,

            "header_state": bytearray(self.header().saveState()).hex()
        }

    def restoreSate(self, state: dict, dataframe: pd.DataFrame):
        if not self.model():
            log.debug(f"{self.__class__.__name__}: Model must first be set on the treeview before using restoreState")
            return

        columns = list(dataframe.columns)

        self.model().beginResetModel()

        self.expanded_paths = set(tuple(p) for p in state.get("expanded_paths", []))
        self.columnFilters = {col: q for col, q in state.get("filters", {}).items() if col in columns}

        self.model().dataframe = dataframe

        self.model().grouped_columns = [columns.index(name) for name in state.get("grouped_columns", []) if name in columns]
        self.model().filtered_columns = {columns.index(name) for name in self.columnFilters if name in columns}

        self.model().sort_column = columns.index(state.get("sort_column")) if state.get("sort_column") in columns else 0
        self.model().sort_order = Qt.SortOrder.AscendingOrder if state.get("sort_ascending") else Qt.SortOrder.DescendingOrder

        self.model()._query = self.buildQuery()

        self.model().endResetModel()

        match = True
        for i, col in enumerate(state.get("columns", [])):
            if i > len(columns) - 1:
                match = False
                break
            if columns[i] != col:
                match = False
                break

        if match:
            self.header().restoreState(bytearray.fromhex(state.get("header_state", "")))

        for i, col in enumerate(columns):
            self.setColumnHidden(i, col not in state.get("visible_columns", [col]))

        self.expand_after_reset()

    def expand_after_reset(self):
        indices = []
        for path in self.expanded_paths:
            try:
                indices.append(self.model().indexFromPath(list(path)))
            except KeyError:
                continue

        for index in indices:
            self.expand(index)




# -*- coding: utf-8 -*-
from collections import namedtuple

from qtpy import QtWidgets

Switches = namedtuple("switches", ("func", "method", "scenario"))


class SwitchComboBox(QtWidgets.QComboBox):
    """For keeping track of contribution tab comparisons."""

    def __init__(self, parent: QtWidgets.QWidget = None):
        super().__init__(parent)
        self.has_scenarios = getattr(parent, "has_scenarios")
        self.switches = Switches("Reference Flows", "Impact Categories", "Scenarios")
        self.indexes = Switches(0, 1, 2)

    def configure(self, has_func: bool = True, has_method: bool = True):
        self.blockSignals(True)
        if all([has_func, has_method]):
            self.insertItems(0, [self.switches.func, self.switches.method])
        if self.has_scenarios:
            self.insertItems(self.indexes.scenario, [self.switches.scenario])
        self.setVisible(self.count() > 0)
        self.blockSignals(False)



import datetime
from logging import getLogger

from qtpy import QtWidgets

import bw2data as bd

from activity_browser import application, signals
from activity_browser.actions.base import ABAction, exception_dialogs

from .project_migrate25 import ProjectMigrate25

log = getLogger(__name__)


class ProjectSwitch(ABAction):
    """
    Switch to a specified Brightway2 project.

    This method compares the given project name with the currently active project.
    If the specified project is different, it switches to the new project, updates
    the last opened timestamp, and logs the change. If the project is not Brightway25
    compatible, a warning is displayed. If the specified project is already active,
    no action is taken.

    Args:
        project_name (str): The name of the project to switch to.

    Logs:
        Warning: If the project is not Brightway25 compatible.
        Info: When the project is successfully switched.
        Debug: If the specified project is already the current project.
    """

    text = "Switch project"
    tool_tip = "Switch the project"

    @staticmethod
    @exception_dialogs
    def run(project_name: str):
        # compare the new to the current project name and switch to the new one if the two are not the same
        if project_name == bd.projects.current:
            log.debug(f"Brightway2 already selected: {project_name}")
            return

        # switch to the new project, don't auto update to brightway25
        bd.projects.set_current(project_name, update=False)

        if not bd.projects.twofive:
            log.warning(f"Project: {bd.projects.current} is not yet BW25 compatible")
            ProjectSwitch.set_warning_bar()

        log.info(f"Brightway2 current project: {project_name}")

        # update the last opened timestamp
        bd.projects.dataset.data["last_opened"] = datetime.datetime.now().isoformat()
        bd.projects.dataset.save()

    @staticmethod
    def set_warning_bar():
        application.main_window.addToolBar(ProjectWarningBar())


class ProjectWarningBar(QtWidgets.QToolBar):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMovable(False)

        warning_label = QtWidgets.QLabel("  This project is not Brightway25 compatible. ")
        height = warning_label.minimumSizeHint().height()

        warning_icon = QtWidgets.QLabel(self)
        qicon = application.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_MessageBoxWarning)
        pixmap = qicon.pixmap(height, height)
        warning_icon.setPixmap(pixmap)

        migrate_label = QtWidgets.QLabel("<a style='text-decoration:underline;'>Migrate project now</a>")
        migrate_label.mouseReleaseEvent = lambda x: ProjectMigrate25.run(bd.projects.current)

        self.addWidget(warning_icon)
        self.addWidget(warning_label)
        self.addWidget(migrate_label)

        signals.project.changed.connect(self.deleteLater)

    def contextMenuEvent(self, event):
        return None




from logging import getLogger

from qtpy.QtWidgets import QMessageBox
from bw2data.proxies import ExchangeProxyBase

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from bw2data.parameters import ActivityParameter
from activity_browser.ui.icons import qicons

from ..parameter.parameter_new_automatic import ParameterNewAutomatic
from .exchange_formula_remove import ExchangeFormulaRemove

log = getLogger(__name__)


class ExchangeModify(ABAction):
    """
    ABAction to modify an exchange with the supplied data.
    """

    icon = qicons.delete
    text = "Modify exchange"

    @classmethod
    @exception_dialogs
    def run(cls, exchange: ExchangeProxyBase, data: dict):
        # remove the formula if it is an empty string
        if "formula" in exchange and data.get("formula") == "":
            del data["formula"]
            ExchangeFormulaRemove.run([exchange])

        for key, value in data.items():
            exchange[key] = value

        exchange.save()

        if "formula" in data:
            cls.parameterize_exchanges(exchange.output.key)

    @staticmethod
    def parameterize_exchanges(key: tuple) -> None:
        """Used whenever a formula is set on an exchange in an activity.

        If no `ActivityParameter` exists for the key, generate one immediately
        """
        act = bd.get_activity(key)
        query = (ActivityParameter.database == key[0]) & (
            ActivityParameter.code == key[1]
        )

        if not ActivityParameter.select().where(query).count():
            ParameterNewAutomatic.run([key])

        group = ActivityParameter.get(query).group

        with bd.parameters.db.atomic():
            bd.parameters.remove_exchanges_from_group(group, act)
            bd.parameters.add_exchanges_to_group(group, act)
            ActivityParameter.recalculate_exchanges(group)



from qtpy import QtCore, QtWidgets

import bw2data as bd
from bw2data.parameters import Group
from bw2data.backends.proxies import ExchangeDataset, Exchanges

from activity_browser import application, settings
from activity_browser.bwutils import AB_metadata
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class DatabaseDelete(ABAction):
    """
    ABAction to delete a database from the project. Asks the user for confirmation. If confirmed, instructs the
    DatabaseController to delete the database in question.
    """

    icon = qicons.delete
    text = "Delete database"
    tool_tip = "Delete this database from the project"

    @staticmethod
    @exception_dialogs
    def run(db_name: str):
        # gathering data will take time for large databases
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

        # get the record count from the database controller
        db_name = db_name
        n_records = AB_metadata.dataframe[AB_metadata.dataframe["database"] == db_name].shape[0]

        # get any upstream exchanges (hacky because Brightway doesn't do this itself)
        excs = Exchanges((db_name, None))
        excs._args = [
            ExchangeDataset.input_database == db_name,
            ExchangeDataset.output_database != db_name,
        ]
        n_upstream_excs = len(excs)

        # construct warning text
        text = f"Are you sure you want to delete database '{db_name}'?"
        if n_records:
            text += f" It contains {n_records} activities"
        if n_upstream_excs:
            text += f" and {n_upstream_excs} exchanges to other databases"

        # ask the user for confirmation
        QtWidgets.QApplication.restoreOverrideCursor()
        response = QtWidgets.QMessageBox.question(
            application.main_window, "Delete database?", text
        )

        # return if the user cancels
        if response != response.Yes:
            return

        # deleting data will take time for large databases
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

        # delete upstream exchanges
        for edge in excs:
            edge.delete()

        # instruct the DatabaseController to delete the database from the project.
        del bd.databases[db_name]

        # delete database search indices
        # fix for: https://github.com/brightway-lca/brightway2-data/issues/219
        # path = os.path.join(bd.projects.request_directory("search"), database.filename)
        # os.remove(path)

        # delete database parameters
        Group.delete().where(Group.name == db_name).execute()

        # remove database from project settings
        settings.project_settings.remove_db(db_name)

        QtWidgets.QApplication.restoreOverrideCursor()



from typing import List

import bw2data as bd
import bw_functional as bf

from activity_browser import bwutils
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class ActivitySDFToClipboard(ABAction):
    """
    ABAction to open one or more supplied activities in an activity tab by employing signals.

    TODO: move away from using signals like this. Probably add a method to the MainWindow to add a panel instead.
    """

    icon = qicons.superstructure
    text = "SDF to clipboard"

    @staticmethod
    @exception_dialogs
    def run(activities: List[tuple | int | bd.Node]):
        activities = [bwutils.refresh_node(node) for node in activities]

        exchanges = []
        for activity in activities:
            if isinstance(activity, bf.Product):
                exchanges += activity.virtual_edges
            if isinstance(activity, bf.Process):
                for product in activity.products():
                    exchanges += product.virtual_exchanges
            else:
                exchanges += [exc.to_dict() for exc in activity.exchanges()]

        df = bwutils.exchanges_to_sdf(exchanges)
        df.to_clipboard(excel=True, index=False)



# -*- coding: utf-8 -*-
import functools
import warnings
from pathlib import Path

import tqdm
from bw2io import BW2Package, ExcelImporter
from bw2io.errors import InvalidPackage, StrategyError
from bw2io.strategies import (assign_only_product_as_production,
                              convert_activity_parameters_to_list,
                              convert_uncertainty_types_to_integers,
                              csv_add_missing_exchanges_section,
                              csv_drop_unknown, csv_numerize,
                              csv_restore_booleans, csv_restore_tuples,
                              drop_falsey_uncertainty_fields_but_keep_zeros,
                              link_iterable_by_fields,
                              link_technosphere_by_activity_hash,
                              normalize_biosphere_categories,
                              normalize_biosphere_names, normalize_units,
                              set_code_by_activity_hash,
                              strip_biosphere_exc_locations)

from activity_browser.mod import bw2data as bd

from .errors import LinkingFailed
from .strategies import (alter_database_name, csv_rewrite_product_key,
                         hash_parameter_group, link_exchanges_without_db,
                         relink_exchanges_bw2package, relink_exchanges_with_db,
                         rename_db_bw2package)


class ABExcelImporter(ExcelImporter):
    """Customized Excel importer for the AB."""

    def write_database(self, **kwargs):
        """Go to the parent of the ExcelImporter class, not the ExcelImporter itself.

        This is important because we want to return a Database instance
        """
        kwargs["activate_parameters"] = kwargs.get("activate_parameters", True)
        return super(ExcelImporter, self).write_database(**kwargs)

    @classmethod
    def simple_automated_import(
        cls, filepath, db_name: str, relink: dict = None
    ) -> list:
        """Handle a lot of the customizable things that can happen
        when doing an import in a script or notebook.
        """

        obj = cls(filepath)
        return obj.automated_import(db_name, relink)

    def automated_import(self, db_name: str, relink: dict = None) -> list:
        self.strategies = [
            functools.partial(alter_database_name, old=self.db_name, new=db_name),
            csv_restore_tuples,
            csv_restore_booleans,
            csv_numerize,
            csv_drop_unknown,
            csv_add_missing_exchanges_section,
            csv_rewrite_product_key,
            normalize_units,
            normalize_biosphere_categories,
            normalize_biosphere_names,
            strip_biosphere_exc_locations,
            set_code_by_activity_hash,
            functools.partial(
                link_iterable_by_fields,
                other=bd.Database(bd.config.biosphere),
                kind="biosphere",
            ),
            assign_only_product_as_production,
            link_technosphere_by_activity_hash,
            drop_falsey_uncertainty_fields_but_keep_zeros,
            convert_uncertainty_types_to_integers,
            hash_parameter_group,
            convert_activity_parameters_to_list,
        ]
        self.db_name = db_name

        # Test if the import contains any parameters.
        has_params = any(
            [
                self.project_parameters,
                self.database_parameters,
                any("parameters" in ds for ds in self.data),
            ]
        )
        self.apply_strategies()
        if any(self.unlinked) and relink:
            for db, new_db in relink.items():
                if db == "(name missing)":
                    self.apply_strategy(
                        functools.partial(link_exchanges_without_db, db=new_db)
                    )
                else:
                    self.apply_strategy(
                        functools.partial(relink_exchanges_with_db, old=db, new=new_db)
                    )
            # Relinking failed (some exchanges still unlinked)
            if any(self.unlinked):
                # Raise a different exception.
                excs = [exc for exc in self.unlinked][:10]
                databases = {
                    exc.get("database", "(name missing)") for exc in self.unlinked
                }
                raise LinkingFailed(excs, databases)
        if any(self.unlinked):
            # Still have unlinked fields? Raise exception.
            excs = [exc for exc in self.unlinked][:10]
            databases = {exc.get("database", "(name missing)") for exc in self.unlinked}
            raise StrategyError(excs, databases)
        if self.project_parameters:
            self.write_project_parameters(delete_existing=False)
        db = self.write_database(delete_existing=True, activate_parameters=True)
        if has_params:
            bd.parameters.recalculate()
        return [db]

    def apply_strategies(self, strategies=None, verbose=False):
        strategies = strategies or self.strategies
        for strategy in tqdm.tqdm(strategies, desc="Applying strategies", total=len(strategies)):
            self.apply_strategy(strategy, verbose)



class ABPackage(BW2Package):
    """Inherits from brightway2 `BW2Package` and handles importing BW2Packages.

    This implementation is done to raise exceptions and show errors on imports
    much faster.
    """
    APPROVED = {
        "bw2calc",
        "bw2data",
        "bw2io",
        "bw2regional",
        "bw2temporalis",
        "bw_functional",
    }

    @classmethod
    def unrestricted_export(cls, obj, path: Path) -> Path:
        """Export a BW2Package outside the project folder."""
        cls._write_file(path, [cls._prepare_obj(obj, False)])
        return path

    @classmethod
    def evaluate_metadata(cls, metadata: dict, ignore_dbs: set):
        """Take the given metadata dictionary and test it against realities
        of the current brightway project.
        """
        if "depends" in metadata:
            missing = set(metadata["depends"]).difference(bd.databases)
            # Remove any databases present in ignore_dbs (these will be relinked)
            missing = missing.difference(ignore_dbs)
            if missing:
                raise InvalidPackage(
                    "Package data links to database names that do not exist: {}".format(
                        missing
                    ),
                    missing,
                )

    @classmethod
    def load_file(cls, filepath, whitelist=True, **kwargs):
        """Similar to how the base class loads the data, but also perform
        a number of evaluations on the metadata.

        Also, if given a 'relink' dictionary, perform relinking of exchanges.
        """
        relink = kwargs.get("relink", None)
        db_name = kwargs.get("rename", None)
        data = super().load_file(filepath, whitelist)
        relinking = set(relink.keys()) if relink else set([])
        if isinstance(data, dict):
            if "metadata" in data:
                cls.evaluate_metadata(data["metadata"], relinking)
            if db_name and "name" in data and data["name"] != db_name:
                old_name = data.pop("name")
                data["name"] = db_name
                data["data"] = rename_db_bw2package(data["data"], old_name, db_name)
            if relink:
                data["data"] = relink_exchanges_bw2package(data["data"], relink)
        else:
            for obj in data:
                if "metadata" in obj:
                    cls.evaluate_metadata(obj["metadata"], relinking)
                if db_name and "name" in obj and obj["name"] != db_name:
                    old_name = obj.pop("name")
                    obj["name"] = db_name
                    obj["data"] = rename_db_bw2package(obj["data"], old_name, db_name)
                if relink:
                    obj["data"] = relink_exchanges_bw2package(obj["data"], relink)
        return data

    @classmethod
    def import_file(cls, filepath, whitelist=True, **kwargs):
        """Import bw2package file, and create the loaded objects, including registering, writing, and processing the created objects.

        Args:
            * *filepath* (str): Path of file to import
            * *whitelist* (bool): Apply whitelist to allowed types. Default is ``True``.
        Kwargs:
            * *relink* (dict): A dictionary of keys with which to relink exchanges
              within the imported package file.

        Returns:
            Created object or list of created objects.

        """
        loaded = cls.load_file(filepath, whitelist, **kwargs)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if isinstance(loaded, dict):
                return cls._create_obj(loaded)
            else:
                return [cls._create_obj(o) for o in loaded]



# -*- coding: utf-8 -*-
from typing import Iterable, Optional

import numpy as np
import pandas as pd
from qtpy.QtWidgets import QPushButton

from activity_browser.mod import bw2data as bd

from ..commontasks import format_activity_label
from ..errors import ScenarioExchangeNotFoundError
from ..multilca import MLCA, Contributions
from ..utils import Index
from .dataframe import (arrays_from_indexed_superstructure,
                        filter_databases_indexed_superstructure,
                        scenario_names_from_df)
from .file_dialogs import ABPopup

try:
    from bw2calc.matrices import TechnosphereBiosphereMatrixBuilder as MB
except ModuleNotFoundError:
    pass  # removed in bw25


class SuperstructureMLCA(MLCA):
    """Subclass of the `MLCA` class which adds another dimension in the form
    of scenarios.
    """

    matrices = {
        "biosphere": "biosphere_matrix",
        "technosphere": "technosphere_matrix",
        "production": "technosphere_matrix",
    }

    def __init__(self, cs_name: str, df: pd.DataFrame):
        assert isinstance(df, pd.DataFrame), (
            "Check if you have provided at least 1 reference flow, 1 impact category "
            "and 1 scenario file. "
        )
        assert not df.empty, "Cannot run analysis without data."
        self.scenario_names = scenario_names_from_df(df)
        self.total = len(self.scenario_names)
        assert self.total > 0, "Cannot run analysis without scenarios"

        super().__init__(cs_name)

        # Scenarios overwrite the lca.xxx_matrix. For supporting absent values
        # in scenario files defaults are required, to prevent these from being
        # overwritten duplicates are required...
        self.default_technosphere_matrix = self.lca.technosphere_matrix.copy()
        self.default_biosphere_matrix = self.lca.biosphere_matrix.copy()
        self.defaults = {
            "technosphere": "default_technosphere_matrix",
            "production": "default_technosphere_matrix",
            "biosphere": "default_biosphere_matrix",
        }

        # Filter dataframe for keys that do not occur in the LCA matrix.
        df = filter_databases_indexed_superstructure(df, self.all_databases)
        assert not df.empty, "Filtering unused flows removed all of the scenario data."

        self.indices, self.values = arrays_from_indexed_superstructure(df)
        # Note: Using the mapping scheme from brightway and presamples,
        # the 'input' keys are matched to the product_dict or
        # biosphere_dict ('rows') while the 'output' keys are matched
        # to the activity_dict ('cols').

        # Side-note on presamples: Presamples was used in AB for calculating scenarios,
        # presamples was superseded by this implementation. For more reading:
        # https://presamples.readthedocs.io/en/latest/use_with_bw2.html
        self.matrix_indices = np.zeros(
            len(self.indices),
            dtype=[
                ("row", np.uint32),
                ("col", np.uint32),
                ("flip", np.bool_),
            ],
        )
        self.indices_to_matrix()

        # Construct an index dictionary similar to fu_index and method_index
        self._current_index = 0
        self.scenario_index = {k: i for i, k in enumerate(self.scenario_names)}

        # Rebuild numpy arrays with scenario dimension included.
        self.lca_scores = np.zeros(
            (len(self.func_units), len(self.methods), self.total)
        )
        self.elementary_flow_contributions = np.zeros(
            (
                len(self.func_units),
                len(self.methods),
                self.total,
                self.lca.biosphere_matrix.shape[0],
            )
        )
        self.process_contributions = np.zeros(
            (
                len(self.func_units),
                len(self.methods),
                self.total,
                self.lca.technosphere_matrix.shape[0],
            )
        )

    @property
    def current(self) -> int:
        return self._current_index

    @current.setter
    def current(self, current: int) -> None:
        """Ensure current index is looped to 0 if end of array is reached."""
        self._current_index = current if current < self.total else 0

    def next_scenario(self):
        self.update_matrices()
        self.current += 1

    def set_scenario(self, index: int) -> None:
        """Set the current scenario index given a new index to go to"""
        steps = self._get_steps_to_index(index)
        # self.current = steps[-1] + 1  # Walk the steps to the new index
        for _ in steps:
            self.next_scenario()

    def indices_to_matrix(self) -> None:
        def convert(idx: Index) -> tuple:
            in_dict = (
                self.lca.biosphere_dict
                if idx.flow_type == "biosphere"
                else self.lca.product_dict
            )
            return (
                in_dict.get(bd.mapping[idx.input]),
                self.lca.activity_dict.get(bd.mapping[idx.output]),
                idx.flip,
            )

        for i, index in enumerate(self.indices):
            try:
                self.matrix_indices[i] = convert(index)
            except (ValueError, KeyError) as e:
                # This is to be used as a fail safe for the case where we don't catch a bad exchange during the import
                # process, or if something else causes an issue with the exchange
                msg = f"One of the activities in the exchange between ({index.input.database}, {index.input.code}) and ({index.output.database}, {index.output.code}) from the scenario file is not present within the designated database. Please check both keys for this exchange within your scenario file with the corresponding databases."
                critical = ABPopup.abCritical(
                    "Scenario Key Error", msg, QPushButton("Cancel")
                )
                critical.exec_()
                raise ScenarioExchangeNotFoundError

    def update_matrices(self) -> None:
        """A Simplified version of the `PackagesDataLoader.update_matrices` method.
        In this case, we expect to only replace technosphere and biosphere
        values, leaving out characterization factor values.
        """
        kinds = set([idx[2] for idx in self.indices])
        types = np.array([idx[2] for idx in self.indices])
        for kind in kinds:
            idx = self.matrix_indices[types == kind]
            sample = self.values[types == kind, self.current]
            # Previously filtered sample and idx for NaN values in samples.
            # Currently replaces sample NaN values with defaults from the databases
            if np.isnan(sample).any():
                default = getattr(self, self.defaults[kind])
                na_idx = idx[np.isnan(sample)]
                if kind == "technosphere":
                    sample[np.isnan(sample)] = np.multiply(
                        default[na_idx["row"], na_idx["col"]].tolist()[0], -1
                    )
                else:
                    sample[np.isnan(sample)] = default[
                        na_idx["row"], na_idx["col"]
                    ].tolist()[0]
            try:
                matrix = getattr(self.lca, self.matrices[kind])
            except AttributeError:
                # This LCA doesn't have this matrix
                continue

            if self.matrices[kind] == "technosphere_matrix":
                # Remove existing matrix factorization
                # because changing technosphere
                if hasattr(self.lca, "solver"):
                    delattr(self.lca, "solver")

            mask = np.where(idx["flip"] == True)
            sample[mask] = -1 * sample[mask]

            matrix[
                idx["row"],
                idx["col"],
            ] = sample

    def _perform_calculations(self):
        """Near copy of `MLCA` class, but includes a loop for all scenarios."""
        for ps_col in range(self.total):
            self.next_scenario()
            for row, func_unit in enumerate(self.func_units):
                try:
                    self.lca.redo_lci(func_unit)
                except:
                    # bw25 compatibility requires activity id instead of activity key
                    key = list(func_unit.keys())[0]
                    self.lca.redo_lci({bd.get_activity(key).id: func_unit[key]})

                self.scaling_factors.update(
                    {(str(func_unit), ps_col): self.lca.supply_array}
                )
                self.technosphere_flows.update(
                    {
                        (str(func_unit), ps_col): np.multiply(
                            self.lca.supply_array,
                            self.lca.technosphere_matrix.diagonal(),
                        )
                    }
                )
                self.inventory.update(
                    {
                        (str(func_unit), ps_col): np.array(
                            self.lca.inventory.sum(axis=1)
                        ).ravel()
                    }
                )
                self.inventories.update({(str(func_unit), ps_col): self.lca.inventory})

                for col, cf_matrix in enumerate(self.method_matrices):
                    self.lca.characterization_matrix = cf_matrix
                    self.lca.lcia_calculation()
                    self.lca_scores[row, col, ps_col] = self.lca.score
                    self.characterized_inventories[(row, col, ps_col)] = (
                        self.lca.characterized_inventory.copy()
                    )
                    self.elementary_flow_contributions[row, col, ps_col] = np.array(
                        self.lca.characterized_inventory.sum(axis=1)
                    ).ravel()
                    self.process_contributions[row, col, ps_col] = (
                        self.lca.characterized_inventory.sum(axis=0)
                    )

    def update_lca_calculation_for_sankey(
        self, scenario_index: int, func_unit: str, method_index: int
    ):
        """
        Reuses the LCA object to prepare the LCA object for necessary calculations to be made before performing the
        Graph Traversal calculations

        @param scenario_index: Index of the Scenario for which the calculation must be performed
        @param func_unit: The functional unit for which the calculation must be performed
        @param method_index: Index of the method for which the calculation must be performed
        """
        self.current = scenario_index
        self.update_matrices()
        try:
            self.lca.redo_lci(func_unit)
        except:
            # brightway25 compatibility
            key = list(func_unit.keys())[0]
            self.lca.redo_lci({bd.get_activity(key).id: func_unit[key]})
        self.lca.characterization_matrix = self.method_matrices[method_index]
        self.lca.lcia_calculation()
        self.lca.decompose_technosphere()

    def get_results_for_method(self, index: int = 0) -> pd.DataFrame:
        """Overrides the parent and returns a dataframe with the scenarios
        as columns
        """
        data = self.lca_scores[:, index, :]
        return pd.DataFrame(data, index=self.func_key_list, columns=self.scenario_names)

    def _get_steps_to_index(self, index: int) -> list:
        """Determine how many steps to take when given the index we want
         to land on.

        We can only iterate through the presample arrays in one direction, so
         if we go from 2 to 1 we need to calculate the amount of steps to loop
         around to 1.
        """
        if index < 0:
            raise ValueError("Negative indexes are not allowed")
        elif index >= self.total:
            raise ValueError("Given index is not possible for current scenario dataset")
        if index < self.current:
            return [*range(self.current, self.total), *range(index)]
        else:
            return list(range(self.current, index))

    def lca_scores_to_dataframe(self) -> pd.DataFrame:
        """Returns a dataframe of LCA scores using FU labels as index and
        the product of methods and scenarios as columns.
        """
        labels = [format_activity_label(k, style="pnld") for k in self.fu_activity_keys]
        methods = [", ".join(m) for m in self.methods]
        df = pd.DataFrame(
            data=[],
            index=pd.Index(labels),
            columns=pd.MultiIndex.from_product(
                [methods, self.scenario_names], names=["method", "scenario"]
            ),
        )
        # Now insert the LCA scores in the correct locations.
        for x, m in enumerate(methods):
            for y, s in enumerate(self.scenario_names):
                idx = pd.MultiIndex.from_tuples([(m, s)])
                df.loc[:, idx] = self.lca_scores[:, x, y]
        return df


class SuperstructureContributions(Contributions):
    mlca: SuperstructureMLCA

    def __init__(self, mlca):
        if not isinstance(mlca, SuperstructureMLCA):
            raise TypeError(
                "Must pass a SuperstructureMLCA object. Passed: {}".format(type(mlca))
            )
        super().__init__(mlca)

    def _build_inventory(
        self, inventory: dict, indices: dict, columns: list, fields: list
    ) -> pd.DataFrame:
        inventory = {k[0]: v for k, v in inventory.items() if k[1] == self.mlca.current}
        return super()._build_inventory(inventory, indices, columns, fields)

    def lca_scores_df(self, normalized: bool = False) -> pd.DataFrame:
        """Returns a metadata-annotated DataFrame of the LCA scores."""
        scores = self.mlca.lca_scores_normalized if normalized else self.mlca.lca_scores
        scores = scores[:, :, self.mlca.current]
        return self._build_lca_scores_df(scores)

    def _build_contributions(
        self, data: np.ndarray, index: int, axis: int
    ) -> np.ndarray:
        data = data[:, :, self.mlca.current]
        return super()._build_contributions(data, index, axis)

    @staticmethod
    def _build_scenario_contributions(
        data: np.ndarray, fu_index: int, m_index: int
    ) -> np.ndarray:
        return data[fu_index, m_index, :]

    def get_contributions(
        self, contribution, functional_unit=None, method=None, scenario=0
    ) -> np.ndarray:
        """Return a contribution matrix given the type and fu / method

        Allow for both fu and method to exist.
        """
        if not any([functional_unit, method]):
            raise ValueError(
                "Either reference flow, impact category or both should be given. Provided:"
                "\n Reference flow: {} \n Impact Category: {}".format(
                    functional_unit, method
                )
            )
        dataset = {
            "process": self.mlca.process_contributions,
            "elementary_flow": self.mlca.elementary_flow_contributions,
        }
        if method and functional_unit:
            return self._build_scenario_contributions(
                dataset[contribution],
                self.mlca.func_key_dict[functional_unit],
                self.mlca.method_index[method],
            )
        self.mlca.current = scenario
        return super().get_contributions(contribution, functional_unit, method)

    def _contribution_index_cols(self, **kwargs) -> (dict, Optional[Iterable]):
        # If both functional_unit and method are given, return scenario index.
        if all(kwargs.values()):
            return self.mlca.scenario_index, self.act_fields
        else:
            return super()._contribution_index_cols(**kwargs)



from qtpy.QtWidgets import QFrame, QWidget


class ABHLine(QFrame):
    def __init__(self, parent: QWidget = None):
        super().__init__(parent)
        self.setFrameShape(QFrame.Shape.HLine)
        self.setFrameShadow(QFrame.Shadow.Sunken)


class ABVLine(QFrame):
    def __init__(self, parent: QWidget = None):
        super().__init__(parent)
        self.setFrameShape(QFrame.Shape.VLine)
        self.setFrameShadow(QFrame.Shadow.Sunken)



# -*- coding: utf-8 -*-
from .checkbox import CheckboxDelegate
from .combobox import ComboBoxDelegate
from .database import DatabaseDelegate
from .delete_button import DeleteButtonDelegate
from .float import FloatDelegate
from .formula import FormulaDelegate
from .json import JSONDelegate
from .list import ListDelegate
from .string import StringDelegate
from .uncertainty import UncertaintyDelegate
from .viewonly import (ViewOnlyDelegate, ViewOnlyFloatDelegate,
                       ViewOnlyUncertaintyDelegate)
from .new_formula import NewFormulaDelegate
from .date_time import DateTimeDelegate
from .property import PropertyDelegate

__all__ = [
    "CheckboxDelegate",
    "ComboBoxDelegate",
    "DatabaseDelegate",
    "DeleteButtonDelegate",
    "FloatDelegate",
    "FormulaDelegate",
    "JSONDelegate",
    "ListDelegate",
    "StringDelegate",
    "UncertaintyDelegate",
    "ViewOnlyDelegate",
    "ViewOnlyFloatDelegate",
    "ViewOnlyUncertaintyDelegate",
    "NewFormulaDelegate",
    "DateTimeDelegate",
    "PropertyDelegate",
]



from typing import List

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils import commontasks
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class ExchangeNew(ABAction):
    """
    ABAction to create a new exchange for an activity.
    """

    icon = qicons.add
    text = "Add exchanges"

    @staticmethod
    @exception_dialogs
    def run(from_keys: List[tuple], to_key: tuple, type: str):
        to_activity = bd.get_activity(to_key)
        for from_key in from_keys:
            exchange = to_activity.new_exchange(input=from_key, type=type, amount=1)
            exchange.save()



# -*- coding: utf-8 -*-
import itertools
from typing import List, Optional, Union
from logging import getLogger

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication, QPushButton

from activity_browser.mod import bw2data as bd

from ..errors import (CriticalScenarioExtensionError, ImportCanceledError,
                      ScenarioExchangeDataNonNumericError,
                      ScenarioExchangeDataNotFoundError,
                      ScenarioExchangeNotFoundError,
                      UnalignableScenarioColumnsWarning)
from .activities import fill_df_keys_with_fields, get_activities_from_keys
from .dataframe import scenario_columns
from .file_dialogs import ABPopup
from .utils import SUPERSTRUCTURE, _time_it_, guess_flow_type

log = getLogger(__name__)

EXCHANGE_KEYS = pd.Index(["from key", "to key"])
INDEX_KEYS = pd.Index(["from key", "to key", "flow type"])


class SuperstructureManager(object):
    """A combination of methods used to manipulate and transform superstructures."""

    def __init__(self, df: pd.DataFrame, *dfs: pd.DataFrame):
        # Prepare dataframes for further processing
        self.frames: List[pd.DataFrame] = [
            SuperstructureManager.format_dataframe(df)
        ] + [SuperstructureManager.format_dataframe(f) for f in dfs]
        self.is_multiple = len(self.frames) > 1

    def combined_data(
        self, kind: str = "product", skip_checks: bool = False
    ) -> pd.DataFrame:
        """
        Combines multiple superstructures using logic specified by the first argument (kind).

        Currently implemented: 'product' creates an outer-product combination
        from all of the columns of the dataframes and injects values from all
        the frames for their specific indexes, any shared indexes are overridden
        where the later dataframes have preference.

        Uses parts of https://stackoverflow.com/a/45286061

        If only a single dataframe is given to the manager, return this dataframe instead.

        Parameters
        ----------
        kind: a string that indicates the type of combination to make for the class self.frames variable
        can be of the form 'product', or 'addition'
        skip_checks: A boolean trigger that should be set to True if the duplicate checks are not required.
        Primarily of use when removing a dataframe, in which case duplicates are removed without warnings
        provided to the user.

        Returns
        -------
        A single pandas dataframe built from the separate dataframes held in the objects frame variable
        """
        if not self.is_multiple:
            df = next(iter(self.frames))
            if skip_checks:
                df = SuperstructureManager.remove_duplicates(df)
            cols = scenario_columns(df)
            SuperstructureManager.check_scenario_exchange_values(df, cols)
            df = SuperstructureManager.merge_flows_to_self(df)
            return pd.DataFrame(data=df.loc[:, cols], index=df.index, columns=cols)
        combo_idx = self._combine_indexes()

        if kind == "product":
            combo_cols = self._combine_columns()
            df = SuperstructureManager.product_combine_frames(
                self.frames, combo_idx, combo_cols, skip_checks
            )
            # Flatten the columns again for later processing.
            df.columns = df.columns.to_flat_index()
        elif kind == "addition":
            # Find the intersection subset of scenarios.
            cols = self._combine_columns_intersect()
            df = SuperstructureManager.addition_combine_frames(
                self.frames, combo_idx, cols, skip_checks
            )
            # Note the dataframe is built with a common index built from all files.
            # So no duplicates will be present in the DataFrame (df), eliminating checks
        else:
            df = pd.DataFrame([], index=combo_idx)
        cols = scenario_columns(df)
        return pd.DataFrame(data=df.loc[:, cols], index=df.index, columns=cols)

    def _combine_columns(self) -> pd.MultiIndex:
        """
        Combines the scenario columns from the objects self.frames variable following combinatoric
        principles.

        Raises
        ------
        CriticalScenarioExtensionError if multiple dataframes in the self.frames variable contain the
        same scenario names

        Returns
        -------
        A pandas multi-index with the separate dataframes in self.frames contributing to the index levels
        """
        cols = [scenario_columns(df).to_list() for df in self.frames]
        return pd.Index([str(c) for c in list(itertools.product(*cols))])

    def _combine_columns_intersect(self) -> pd.Index:
        iterable = iter(self.frames)
        cols = scenario_columns(next(iterable))
        absent = set()
        for df in iterable:
            absent.update(cols.symmetric_difference(scenario_columns(df)))
            cols = cols.intersection(scenario_columns(df))
            for name in absent:
                log.warning(
                    "The following scenario is not found in all provided files and is being dropped: {}".format(
                        name
                    )
                )
        if cols.empty:
            msg = "While attempting to combine the scenario files an error was detected. No scenario columns were found in common between the files. For combining scenarios by extension at least one scenario needs to be found in common."
            critical = ABPopup.abCritical(
                "Combining scenario files.", msg, QPushButton("Cancel")
            )
            critical.exec_()
            raise CriticalScenarioExtensionError
        elif len(absent) > 0:
            msg = (
                "<p>While importing the scenario difference files one, or more, of the scenarios could not be found "
                "between the files.</p> In these circumstances the Activity-Browser will only retain those "
                "scenarios found in common between these files. If some desired scenarios are not included, then "
                "please inspect your scenario files for the relevant columns."
            )
            warning = ABPopup.abWarning(
                "Scenarios being dropped", msg, QPushButton("Ok"), QPushButton("Cancel")
            )
            warning.dataframe(pd.DataFrame({"Scenarios": list(absent)}), ["Scenarios"])
            response = warning.exec_()
            if response == warning.DialogCode.Rejected:
                raise UnalignableScenarioColumnsWarning()
        return cols

    def _combine_indexes(self) -> pd.MultiIndex:
        """Returns a union of all of the given dataframe indexes."""
        iterable = iter(self.frames)
        idx = next(iterable).index
        for df in iterable:
            idx = idx.union(df.index)
        return idx

    @staticmethod
    def product_combine_frames(
        data: List[pd.DataFrame],
        index: pd.MultiIndex,
        cols: pd.MultiIndex,
        skip_checks: bool = False,
    ) -> pd.DataFrame:
        """Iterate through the dataframes, filling data into the combined
        dataframe with duplicate indexes being resolved using a 'last one wins'
        logic.

        Parameters
        ----------
        data: A List of dataframes, each dataframe corresponding to a dataframe from a single scenario difference file
        index: The combined Multi-index for the final merged dataframe
        cols: A Multi-index object that contains all of the levels (scenario names) from the different scenario files)
        this is used to create a combined set of scenario names
        skip_checks: a boolean that triggers the use of duplicate checks (not required when removing scenario files)

        Returns
        -------
        A pandas dataframe constructed from the combined inputs to the class self.frames variable
        """

        def combine(one, two):
            """Should hopefully provide a failsafe approach to combining the different scenario combinations,
            by using a simple vector - vector assignment approach.
            """
            for col_two in SUPERSTRUCTURE.symmetric_difference(two.columns):
                for idx in one.columns:
                    if col_two in idx:
                        one.loc[two.index, idx] = two.loc[:, col_two]

        base_scenario_data = pd.DataFrame([], index=index, columns=SUPERSTRUCTURE)
        scenarios_data = pd.DataFrame([], index=index, columns=cols)
        if not skip_checks:
            tmp_df = SuperstructureManager.check_duplicates(data)
            for idx, f in enumerate(tmp_df):
                SuperstructureManager.check_scenario_exchange_values(
                    f, scenario_columns(f)
                )
                combine(scenarios_data, f)
                base_scenario_data.loc[f.index, :] = f.loc[:, SUPERSTRUCTURE]
        else:
            for idx, f in enumerate(data):
                f = SuperstructureManager.remove_duplicates(f)
                combine(scenarios_data, f)
                base_scenario_data.loc[f.index, :] = f.loc[:, SUPERSTRUCTURE]

        scenarios_data.columns = cols.to_flat_index()
        df = pd.concat([base_scenario_data, scenarios_data], axis=1)
        df = SuperstructureManager.merge_flows_to_self(df)
        #        df.replace(np.nan, 0, inplace=True)
        return df

    @staticmethod
    def addition_combine_frames(
        data: List[pd.DataFrame],
        index: pd.MultiIndex,
        cols: pd.Index,
        skip_checks: bool = False,
    ) -> pd.DataFrame:
        """
        Iterates through the combined dataframes to produce a single merged dataframe where duplicates are resolved
        with a "last one wins" approach

        Parameters
        ----------
        data: list of input dataframes from the scenario difference files
        index: pandas Multi-Index used to create the x-index from the merged indexes of the data input
        cols: the list of scenario columns from the scenario difference files
        skip_checks: a simple boolean used to avoid the duplicate checks (when removing a scenario file)

        Returns
        -------
        A pandas dataframe constructed from the combined inputs to the class self.frames variable
        """
        #        columns = data.columns if isinstance(data, pd.DataFrame) else data[0].columns
        columns = SUPERSTRUCTURE.append(cols)
        df = pd.DataFrame([], index=index, columns=columns)
        if not skip_checks:
            SuperstructureManager.check_duplicates(data)
            for f in data:
                SuperstructureManager.check_scenario_exchange_values(f, cols)
                df.loc[f.index, columns] = f.loc[:, columns]
        else:
            for f in data:
                f = SuperstructureManager.remove_duplicates(f)
                df.loc[f.index, columns] = f.loc[:, columns]
        df = SuperstructureManager.merge_flows_to_self(df)
        #        df.replace(np.nan, 0, inplace=True)
        return df.loc[:, cols]

    @staticmethod
    def format_dataframe(df: pd.DataFrame) -> pd.DataFrame:
        """Format the input superstructure dataframes.
        If in the future more formatting functions are needed, they should be added here.
        """
        if not isinstance(df.index, pd.MultiIndex):
            df.index = SuperstructureManager.build_index(df)
        # all import checks should take place before merge_flows_to_self
        #        df = SuperstructureManager.check_duplicates(df)
        #        df = SuperstructureManager.merge_flows_to_self(df)

        return df

    @staticmethod
    @_time_it_
    def merge_flows_to_self(df: pd.DataFrame) -> pd.DataFrame:
        """
        This function checks if any technosphere flows to self exist and merges them with a production flow.
        If no production flow exists, it is added using the default value from the respective brightway database

        Parameters
        ----------
        df: a pandas dataframe for the scenario files with the exchanges to be checked

        Returns
        -------
        A pandas dataframe with the changes made to the scenario dataframe for these self referential flows
        """
        self_referential_production_flows = df.loc[
            df.apply(
                lambda x: (
                    True
                    if x["from key"] == x["to key"] and x["flow type"] == "technosphere"
                    else False
                ),
                axis=1,
            ),
            :,
        ].copy()
        self_referential_production_flows.index = pd.MultiIndex.from_arrays(
            [
                self_referential_production_flows.index.get_level_values(0),
                self_referential_production_flows.index.get_level_values(1),
                self_referential_production_flows.index.get_level_values(2).str.replace(
                    "technosphere", "production"
                ),
            ],
            names=["input", "output", "flow"],
        )
        scenario_cols = df.columns.difference(SUPERSTRUCTURE)
        prod_indexes = self_referential_production_flows.loc[
            self_referential_production_flows.index.isin(df.index)
        ].index
        self_referential_production_flows.loc[prod_indexes, scenario_cols] = df.loc[
            df.index.isin(self_referential_production_flows.index), scenario_cols
        ]
        self_referential_production_flows.loc[prod_indexes, "flow type"] = "production"

        # TODO use metadata for the default production values
        for idx in self_referential_production_flows.loc[
            ~self_referential_production_flows.index.isin(df.index)
        ].index:

            # this flow to self does not have a similar 'production' flow to self.
            # find the default production value and add it as a 'production' flow

            # WARNING: this way of getting the production amount only works for processes with
            # 1 reference flow (because we just take index 0 from list of production exchanges)
            # Once AB has support for multiple reference flows, we need to adjust this code to match the
            # right flow -something with looping over the flows and getting the right product or something-.
            prod_amt = list(bd.get_activity(idx[0]).production())[0].get("amount", 1)
            # make a new df to edit the production, add the correct values/indices where needed
            # and concat to the main df
            self_referential_production_flows.loc[idx, ["flow type"]] = "production"
            self_referential_production_flows.loc[idx, scenario_cols] = prod_amt
        if len(self_referential_production_flows) > 0:
            tech_idxs = [
                (x[0], x[1], "technosphere")
                for x in self_referential_production_flows.index
            ]

            denominator = (
                self_referential_production_flows.loc[:, scenario_cols]
                + df.loc[tech_idxs, scenario_cols].values
            )
            self_referential_production_flows.loc[:, scenario_cols] = (
                self_referential_production_flows.loc[:, scenario_cols] / denominator
            )
            # if we did divide by 0 then replace these nans by 0
            self_referential_production_flows.loc[:, scenario_cols] = (
                self_referential_production_flows.loc[:, scenario_cols].where(
                    ~denominator.isin([0]), 0
                )
            )

            # drop the 'technosphere' flows
            df = df.drop(tech_idxs)
            df = df.drop(prod_indexes)
            df = pd.concat([df, self_referential_production_flows], axis=0)
        return df

    @staticmethod
    def remove_duplicates(df: pd.DataFrame) -> pd.DataFrame:
        """Using the input/output index for a superstructure, drop duplicates
        where the last instance survives.
        """
        duplicates = df.index.duplicated(keep="last")
        if duplicates.any():
            log.warning(
                "Found and dropped {} duplicate exchanges.".format(duplicates.sum())
            )
            return df.loc[~duplicates, :]
        return df

    @staticmethod
    def build_index(df: pd.DataFrame) -> pd.MultiIndex:
        """Construct MultiIndex from exchange keys and flows, allowing for
        data merging.

        - If any of the exchange key columns are missing keys, attempt to fill
        them. If filling them does not succeed, raise an assertion.
        """
        unknown_flows = df.loc[:, "flow type"].isna()
        if unknown_flows.any():
            log.warning(
                "Not all flow types are known, guessing {} flows".format(
                    unknown_flows.sum()
                )
            )
            df.loc[unknown_flows, "flow type"] = df.loc[
                unknown_flows, EXCHANGE_KEYS
            ].apply(guess_flow_type, axis=1)
        return pd.MultiIndex.from_tuples(
            df.loc[:, INDEX_KEYS].apply(tuple, axis=1),
            names=["input", "output", "flow"],
        )

    @staticmethod
    def exchangesPopup() -> ABPopup:
        """
        Provides a popup message if there is an issue to find some of the scenario exchanges in the uploaded files

        Returns
        -------
        A QDialog with a critical Error
        """
        msg = (
            "<p>One, or several, exchanges (rows) in the scenario file could not be found in the database (meaning:"
            " a part or all of the exchange information, i.e. input or output product/activity/unit/geography, or the"
            " key, have no match in the project databases).</p> <p>It is not possible to proceed at this point."
            " you may save the scenario file with an additional column indicating the problematic exchanges.</p>"
        )
        pop = ABPopup.abCritical(
            "Exchange(s) not found", msg, QPushButton("Save"), QPushButton("Cancel")
        )
        pop.save_options()
        return pop

    @staticmethod
    @_time_it_
    def fill_empty_process_keys_in_exchanges(df: pd.DataFrame) -> pd.DataFrame:
        """identifies those exchanges in the input dataframe that are missing keys.
        If the keys cannot be found in the available databases then an Exception is
        raised

        Raises
        ------
        ScenarioExchangeNotFoundError if a scenario exchange is not present in the local database

        Parameters
        ----------
        df: the input dataframe containing scenario data with exchanges that need to be
        checked for the presence of a key

        Returns
        -------
        A pandas dataframe with complete entries for the dataframes 'to key' and 'from key' fields
        """
        if df.loc[:, EXCHANGE_KEYS].isna().any().any():
            df = fill_df_keys_with_fields(df)
            _df = df.loc[df.loc[:, EXCHANGE_KEYS].isna().any(axis=1)]
            if not _df.empty:
                sdf_keys = SuperstructureManager.exchangesPopup()
                sdf_keys.dataframe_to_file(df, _df.index)
                QApplication.restoreOverrideCursor()
                sdf_keys.exec_()
                raise ScenarioExchangeNotFoundError(
                    "Cannot find key(s) in local databases."
                )
        return df

    @staticmethod
    @_time_it_
    def verify_scenario_process_keys(df: pd.DataFrame) -> pd.DataFrame:
        """Checks all process keys in the scenario file and does not provide alternative keys based on exchange
        metadata.

        Raises
        ------
        ScenarioExchangeNotFoundError if a scenario exchange is not present in the local database

        Parameters
        -------
        df: the dataframe with process keys that need to be verified

        Returns
        -------
        A scenario dataframe with all scenario exchange keys verified with the local brightway databases
        """
        dbs = set(df.loc[:, "from database"]).union(df.loc[:, "to database"])
        df_ = pd.DataFrame({}, columns=df.columns)
        for db in dbs:
            _ = get_activities_from_keys(df, db)
            df_ = pd.concat([df_, _], axis=0, ignore_index=False)
        if not df_.empty:
            sdf_keys = SuperstructureManager.exchangesPopup()
            sdf_keys.dataframe(df_, SUPERSTRUCTURE)
            sdf_keys.dataframe_to_file(df, df_.index)
            QApplication.restoreOverrideCursor()
            sdf_keys.exec_()
            raise ScenarioExchangeNotFoundError(
                "A key provided in the scenario file is not valid for the available database, consult the respective output."
            )

    @staticmethod
    @_time_it_
    def check_scenario_exchange_values(df: pd.DataFrame, cols: pd.Index):
        """ "
        Checks the scenario exchange amounts from the dataframes for valid values, if none are found an error
        is raised, if some exchange amounts are absent a warning is raised and default values are used.

        Raises
        ------
        A ScenarioExchangeDataNotFoundError if no valid values are found in the scenario 'amounts'
        A ScenarioExchangeDataNonNumericError if non-numeric values are found for the scenario 'amounts'
        A logged warning before replacement of invalid scenario values

        Parameters
        ----------
        df: a pandas dataframe holding the current file scenario data, should be in the full scenario file format, with
        all fields defined in the utils.SUPERSTRUCTURE global
        cols: a pandas index that indicates the scenario columns holding the 'amounts' to be used in the scenario
        calculations
        """
        _df = df.copy()
        assert len(cols) > 0
        nas = _df.loc[:, cols].isna()
        if nas.all(axis=0).all():
            msg = (
                "<p>No exchange values could be observed in the last loaded scenario file. "
                + "Exchange values must be recorded in a labelled scenario column with a name distinguishable from the"
                + " default (required) columns, which are:</p>"
                + SuperstructureManager.edit_superstructure_for_string()
                + "<p>Please check the file contents for the scenario columns and the exchange amounts before loading again.</p>"
            )
            critical = ABPopup.abCritical(
                "No scenario exchange data", msg, QPushButton("Cancel")
            )
            critical.exec_()
            raise ScenarioExchangeDataNotFoundError
        elif nas.any(axis=0).any():
            log.warning(
                "Replacing empty values from the last loaded scenario difference file"
            )
        if not is_numeric_dtype(np.array(_df.loc[:, cols])):
            # converting to numeric only works on lists and with the coercive option
            # any errors convert to np.nan and can then only be excluded if previous
            # NaNs are masked by conversion to numeric values
            _df.loc[:, cols].fillna(0, inplace=True)
            bad_entries = pd.DataFrame(index=_df.index)
            for col in cols:
                bad_entries[col] = pd.to_numeric(df.loc[:, col], errors="coerce")
            msg = (
                "<p>Non-numeric data is present in the scenario exchange columns.</p><p> The Activity-Browser can "
                "only deal with numeric data for the calculations. To resolve this corrections will need to be made "
                "to these values in the scenario file.</p>"
            )
            critical = ABPopup.abCritical(
                "Bad (non-numeric) input data",
                msg,
                QPushButton("Save"),
                QPushButton("Cancel"),
            )
            QApplication.restoreOverrideCursor()
            critical.dataframe(df[bad_entries.isna().any(axis=1)], SUPERSTRUCTURE)
            critical.save_options()
            critical.dataframe_to_file(df, bad_entries.isna().any(axis=1))
            critical.exec_()
            raise ScenarioExchangeDataNonNumericError()

    @staticmethod
    @_time_it_
    def check_duplicates(
        data: Optional[Union[pd.DataFrame, list]],
        index: list = ["to key", "from key", "flow type"],
    ):
        """
        Checks three fields to identify whether a scenario difference file contains duplicate exchanges:
        'from key', 'to key' and 'flow type'
        Produces a warning

        Raises
        ------
        ImportCanceledError if the user cancels the import due to duplicate exchanges

        Parameters
        ----------
        data: a pandas dataframe or list of pandas dataframes that will be checked for containing duplicates
        index: nominally required, but probably best to avoid overwriting the default values. Used to indicate the
            columns to check for duplication of an exchange

        Returns
        -------
        A dataframe that contains only unique flow exchanges
        """
        if isinstance(data, pd.DataFrame):
            return SuperstructureManager._check_duplicate(data, index)
        else:
            # Each time the frames are gathered into a list
            # and we are always checking the last file
            # So only comparisons with the last file are required
            count = 1
            df = data[-count].copy()
            duplicated = {}
            while count < len(data):
                count += 1
                popped = data[-count].copy()
                duplicates = SuperstructureManager._check_duplicates(df, popped, count)
                if not duplicates.empty:
                    duplicated[count] = duplicates

            if duplicated:
                msg = (
                    "<p>Duplicates have been found, meaning that there are several rows in the scenario file describing "
                    "scenarios for the same flow. The AB can deal with this by discarding all but the last row for this "
                    "exchange.</p> <p>Press 'Ok' to proceed, press 'Cancel' to abort.</p>"
                )
                warning = ABPopup.abWarning(
                    "Duplicate flow exchanges",
                    msg,
                    QPushButton("Ok"),
                    QPushButton("Cancel"),
                )
                warning.dataframe(
                    pd.concat([file for file in duplicated.values()]),
                    ["file"] + SUPERSTRUCTURE.tolist(),
                )
                QApplication.restoreOverrideCursor()
                response = warning.exec_()
                QApplication.setOverrideCursor(Qt.WaitCursor)
                if response == warning.DialogCode.Rejected:
                    raise ImportCanceledError
            return data

    @staticmethod
    def _check_duplicates(
        dfp: pd.DataFrame,
        pdf: pd.DataFrame,
        count: int,
        index: list = ["to key", "from key", "flow type"],
    ) -> pd.DataFrame:
        # TODO fix variable names 'dfp' & 'pdf' to something undearstandeable
        # TODO create useful docstring, already clear this is a private method from '_' prefix
        """NOT TO BE USED OUTSIDE OF CALLING METHOD check_duplicates"""
        # First save the original index and create a new one that can help the user identify duplicates in their files
        d_idx = dfp.index
        dfp.insert(0, "file", "1", allow_duplicates=True)  # add the file number
        dfp.index = pd.Index([str(i) for i in range(dfp.shape[0])])
        p_idx = pdf.index
        pdf.insert(0, "file", str(count), allow_duplicates=True)  # add the file number
        pdf.index = pd.Index([str(i) for i in range(pdf.shape[0])])
        df = pd.concat([dfp, pdf], ignore_index=True)
        dfp.index = d_idx
        pdf.index = p_idx
        dfp.drop_duplicates(index, keep="last", inplace=True)
        return df.loc[df.duplicated(index, keep=False)]

    @staticmethod
    def _check_duplicate(
        data: pd.DataFrame, index: list = ["to key", "from key", "flow type"]
    ) -> pd.DataFrame:
        # TODO create useful docstring, already clear this is a private method from '_' prefix
        """NOT TO BE USED OUTSIDE OF CALLING METHOD check_duplicates"""
        df = data.copy()
        df.index = pd.Index([str(i) for i in range(df.shape[0])])
        duplicates = df.duplicated(index, keep=False)
        if duplicates.any():
            msg = (
                "<p>Duplicates have been found, meaning that there are several rows in the scenario file describing "
                "scenarios for the same flow. The AB can deal with this by discarding all but the last row for this "
                "exchange.</p> <p>Press 'Ok' to proceed, press 'Cancel' to abort.</p>"
            )
            warning = ABPopup.abWarning(
                "Duplicate flow exchanges",
                msg,
                QPushButton("Ok"),
                QPushButton("Cancel"),
            )
            warning.dataframe(df.loc[duplicates], SUPERSTRUCTURE)
            QApplication.restoreOverrideCursor()
            response = warning.exec_()
            QApplication.setOverrideCursor(Qt.WaitCursor)
            if response == warning.DialogCode.Rejected:
                raise ImportCanceledError
            data.drop_duplicates(index, keep="last", inplace=True)
        return data



from typing import Callable, List, Union

from qtpy import QtCore

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils import commontasks
from bw2data import get_activity
from activity_browser.ui.icons import qicons


class ActivityDuplicate(ABAction):
    """
    Duplicate one or multiple activities using their keys. Proxy action to call the controller.
    """

    icon = qicons.copy
    text = "Duplicate ***"

    @staticmethod
    @exception_dialogs
    def run(activity_keys: List[tuple]):
        activities = [get_activity(key) for key in activity_keys]

        for activity in activities:
            new_code = commontasks.generate_copy_code(activity.key)
            activity.copy(new_code)



import pandas as pd
from qtpy import QtGui, QtCore


class ABAbstractItem:

    def __init__(self, key, parent=None):
        self._key = key
        self._child_keys = []
        self._child_items = {}
        self._parent = None

        if parent:
            self.set_parent(parent)

    def __getitem__(self, item):
        raise NotImplementedError

    def parent(self) -> "ABAbstractItem":
        return self._parent

    def key(self):
        return self._key

    def children(self):
        return self._child_items

    def path(self) -> [str]:
        return self.parent().path() + [self.key()] if self.parent() else []

    def rank(self) -> int:
        """Return the rank of the ABItem within the parent. Returns -1 if there is no parent."""
        if self.parent is None:
            return -1
        return self.parent()._child_keys.index(self.key())

    def has_children(self) -> bool:
        return bool(self._child_keys)

    def set_parent(self, parent: "ABAbstractItem"):
        if self.key() in parent.children():
            raise KeyError(f"Item {self.key()} is already a child of {parent.key()}")

        if self.parent():
            self.parent()._child_keys.remove(self.key())
            del self.parent()._child_items[self.key()]

        parent._child_items[self.key()] = self
        parent._child_keys.append(self.key())
        self._parent = parent

    def loc(self, key_or_path: object | list[object], default=None):
        key = key_or_path.pop(0) if isinstance(key_or_path, list) else key_or_path

        if isinstance(key_or_path, list) and len(key_or_path) > 0:
            return self._child_items[key].loc(key_or_path, default)

        return self._child_items.get(key, default)

    def iloc(self, index: int, default=None):
        return self.loc(self._child_keys[index], default)

    def displayData(self, col: int, key: str):
        return None

    def decorationData(self, col: int, key: str):
        return None

    def fontData(self, col: int, key: str):
        return None

    def backgroundData(self, col: int, key: str):
        return None

    def foregroundData(self, col: int, key: str):
        return None

    def flags(self, col: int, key: str):
        return QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable

    def setData(self, col: int, key: str, value):
        return False


class ABBranchItem(ABAbstractItem):

    def __getitem__(self, item):
        return None

    def put(self, item: ABAbstractItem, path):
        key = path.pop(0)
        if path:
            sub = self.loc(key)
            sub = sub if sub else self.__class__(key, self)
            sub.put(item, path)
        else:
            item.set_parent(self)

    def set_parent(self, parent: "ABAbstractItem"):
        if self.key() in parent._child_items:
            twin = parent.loc(self.key())
            for child in twin.child_items.values():
                child.set_parent(self)

        if self.parent():
            self.parent()._child_keys.remove(self.key())
            del self.parent()._child_items[self.key()]

        parent._child_items[self.key()] = self

        branches = [isinstance(parent._child_items[key], ABBranchItem) for key in parent._child_keys]
        i = branches.index(False) if False in branches else len(branches)
        parent._child_keys.insert(i, self.key())
        self._parent = parent

    def displayData(self, col: int, key: str):
        if col == 0:
            return self.key()
        else:
            return None


class ABDataItem(ABAbstractItem):
    def __init__(self, key, data, parent=None):
        super().__init__(key, parent)
        self.data = data

    def __getitem__(self, item):
        return self.data.get(item)

    def displayData(self, col: int, key: str):
        data = self[key]

        if data is None or pd.isna(data):
            return None

        if isinstance(data, str):
            # clean up the data to a table-readable format
            data = data.replace("\n", " ")

        return data

    def fontData(self, col: int, key: str):
        font = QtGui.QFont()

        # set the font to italic if the display value is Undefined
        if self.displayData(col, key) == "Undefined":
            font.setItalic(True)

        return font



from qtpy import QtWidgets, QtCore
from logging import getLogger

import bw2data as bd
from bw2io import backup

from activity_browser import application, utils
from activity_browser.actions.base import ABAction, exception_dialogs

from activity_browser.ui.threading import ABThread
from activity_browser.ui.icons import qicons

log = getLogger(__name__)


class ProjectNewFromTemplate(ABAction):
    """
    ABAction to create a new project from a remote template.
    """

    icon = qicons.add
    text = "New project from remote"
    tool_tip = "Make a new project from remote template"

    @staticmethod
    @exception_dialogs
    def run(template_key: str):

        if template_key not in utils.get_templates():
            raise ValueError(f"Template key '{template_key}' not found.")

        name, ok = QtWidgets.QInputDialog.getText(
            application.main_window,
            "Create project from template",
            "Name of new project:" + " " * 25,
        )

        if not ok or not name:
            return

        if name in bd.projects:
            QtWidgets.QMessageBox.information(
                application.main_window,
                "Not possible.",
                "A project with this name already exists.",
            )
            return

        # setup dialog
        progress = QtWidgets.QProgressDialog(
            parent=application.main_window,
            labelText="Creating project from template",
            maximum=0
        )
        progress.setCancelButton(None)
        progress.setWindowTitle("Creating project from template")
        progress.setWindowFlag(QtCore.Qt.WindowContextHelpButtonHint, False)
        progress.setWindowFlag(QtCore.Qt.WindowCloseButtonHint, False)
        progress.findChild(QtWidgets.QProgressBar).setTextVisible(False)
        progress.resize(400, 100)
        progress.show()

        # setup the import
        thread = ImportThread(application)
        setattr(thread, "path", utils.get_templates()[template_key])
        setattr(thread, "project_name", name)

        thread.finished.connect(lambda: progress.deleteLater())
        thread.finished.connect(lambda: bd.projects.set_current(name))

        # start the import
        thread.start()


class ImportThread(ABThread):
    path: str
    project_name: str

    def run_safely(self):
        log.debug('Creating project from template:'
                  f'\nPATH: {self.path}'
                  f'\nNAME: {self.project_name}')
        backup.restore_project_directory(fp=self.path, project_name=self.project_name)
        log.info(f"Project `{self.project_name}` created.")





from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons
from activity_browser.ui.threading import ABThread


class DatabaseDuplicate(ABAction):
    """
    ABAction to duplicate a database. Asks the user to provide a new name for the database, and returns when the name is
    already in use by an existing database. Then it shows a progress dialogue which will construct a new thread in which
    the database duplication will take place. This thread instructs the DatabaseController to duplicate the selected
    database with the chosen name.
    """

    icon = qicons.duplicate_database
    text = "Duplicate database..."
    tool_tip = "Make a duplicate of this database"

    @staticmethod
    @exception_dialogs
    def run(db_name: str):
        assert db_name in bd.databases

        new_name, ok = QtWidgets.QInputDialog.getText(
            application.main_window,
            f"Copy {db_name}",
            "Name of new database:" + " " * 25,
        )
        if not new_name or not ok:
            return

        if new_name in bd.databases:
            QtWidgets.QMessageBox.information(
                application.main_window,
                "Not possible",
                "A database with this name already exists.",
            )
            return

        DuplicateDatabaseDialog(db_name, new_name, application.main_window)


class DuplicateDatabaseDialog(QtWidgets.QProgressDialog):
    def __init__(self, from_db: str, to_db: str, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle("Duplicating database")
        self.setLabelText(
            f"Duplicating existing database <b>{from_db}</b> to new database <b>{to_db}</b>:"
        )
        self.setModal(True)
        self.setRange(0, 0)

        self.dup_thread = DuplicateDatabaseThread(from_db, to_db, self)
        self.dup_thread.finished.connect(self.thread_finished)

        self.show()

        self.dup_thread.start()

    def thread_finished(self) -> None:
        self.dup_thread.exit(0)
        self.setMaximum(1)
        self.setValue(1)


class DuplicateDatabaseThread(ABThread):
    def __init__(self, from_db, to_db, parent=None):
        super().__init__(parent=parent)
        self.copy_from = from_db
        self.copy_to = to_db

    def run_safely(self):
        database = bd.Database(self.copy_from)
        database.copy(self.copy_to)



# -*- coding: utf-8 -*-
"""
bwutils is a collection of methods that build upon brightway2 and are generic enough to provide here so that we avoid
re-typing the same code in different parts of the Activity Browser.
"""
import bw_functional

from .commontasks import cleanup_deleted_bw_projects as cleanup
from .commontasks import (refresh_node, refresh_node_or_none, refresh_parameter, refresh_edge, refresh_edge_or_none,
                          parameters_in_scope, exchanges_to_sdf, database_is_locked, database_is_legacy, projects_by_last_opened)
from .metadata import AB_metadata
from .montecarlo import MonteCarloLCA
from .multilca import MLCA, Contributions
from .pedigree import PedigreeMatrix
from .sensitivity_analysis import GlobalSensitivityAnalysis
from .superstructure import SuperstructureContributions, SuperstructureMLCA
from .uncertainty import (CFUncertaintyInterface, ExchangeUncertaintyInterface,
                          ParameterUncertaintyInterface,
                          get_uncertainty_interface)
from .utils import Parameter



from logging import getLogger
from copy import deepcopy

import requests

import ecoinvent_interface as ei
import bw2data as bd

from qtpy import QtWidgets, QtCore
from qtpy.QtCore import Signal, SignalInstance

from activity_browser import application, signals
from activity_browser.ui import widgets, icons, threading
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils.io.ecoinvent_importer import Ecoinvent7zImporter
from activity_browser.bwutils.io.ecoinvent_lcia_importer import EcoinventLCIAImporter

log = getLogger(__name__)


class DatabaseImportFromEcoinvent(ABAction):
    """ABAction to open the DatabaseImportWizard"""

    icon = icons.qicons.import_db
    text = "Import database from ecoinvent"
    tool_tip = "Import database from ecoinvent"

    @staticmethod
    @exception_dialogs
    def run():
        """Show the setup dialog in which the user can first login and then choose the name of the database
        and whether to import the accompanied biosphere, or connect to an existing biosphere."""
        setup = EiWizard(application.main_window)
        setup.exec_()


class EiWizard(widgets.ABWizard):
    """Wizard for importing database from ecoinvent"""

    class RemoteOrLocalPage(widgets.ABWizardPage):
        """Wizard page to choose between remote or local ecoinvent release"""
        title = "Import from ecoinvent"
        subtitle = "Choose whether to import from a remote or local ecoinvent release."

        def __init__(self, parent=None):
            super().__init__(parent)

            self.remote_button = QtWidgets.QRadioButton("Remote")
            self.local_button = QtWidgets.QRadioButton("Local")
            self.remote_button.setChecked(True)

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.remote_button)
            layout.addWidget(self.local_button)
            self.setLayout(layout)

        def nextPage(self):
            """Determine the next page based on the user's selection"""
            if self.local_button.isChecked():
                return EiWizard.LocalSelectPage
            else:
                return EiWizard.LoginPage

    class LocalSelectPage(widgets.ABWizardPage):
        """Wizard page to select a local ecoinvent .7z file"""
        title = "Import from ecoinvent"
        subtitle = "Select local ecoinvent .7z."

        def __init__(self, parent=None):
            super().__init__(parent)

            self.file_selector = widgets.ABFileSelector(filter="*.7z")
            self.file_selector.textChanged.connect(lambda: self.completeChanged.emit())

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.file_selector)
            self.setLayout(layout)

        def finalize(self, context: dict):
            """Store the selected file path in the context"""
            context["ei_filepath"] = self.file_selector.text()

        def isComplete(self):
            """Check if a file has been selected"""
            return bool(self.file_selector.text())

        def nextPage(self):
            """Proceed to the BiosphereSetupPage"""
            return EiWizard.BiosphereSetupPage

    class LoginPage(widgets.ABWizardPage):
        """Wizard page to login with ecoinvent credentials"""
        title = "Login"
        subtitle = "Login with your ecoinvent credentials to authorize the download"

        def __init__(self, parent=None):
            super().__init__(parent)

            self.release = None

            self.username = QtWidgets.QLineEdit()
            self.username.setPlaceholderText('ecoinvent username')

            self.password = QtWidgets.QLineEdit()
            self.password.setPlaceholderText('ecoinvent password')
            self.password.setEchoMode(QtWidgets.QLineEdit.Password)

            self.message = QtWidgets.QLabel()

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.username)
            layout.addWidget(self.password)
            layout.addWidget(self.message)

            self.setLayout(layout)

        def initializePage(self, context: dict):
            """Initialize the page with stored username and password"""
            settings = ei.Settings()
            self.username.setText(settings.username)
            self.password.setText(settings.password)

        def validatePage(self):
            """Validate the login credentials by attempting to list ecoinvent versions"""
            QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
            try:
                settings = ei.Settings(username=self.username.text(), password=self.password.text())
                self.release = ei.EcoinventRelease(settings)
                self.release.list_versions()
            except requests.exceptions.HTTPError as e:
                QtWidgets.QApplication.restoreOverrideCursor()
                if e.response.status_code == 401:
                    self.message.setText("Invalid username and/or password, please try again.")
                    return False
                else:
                    self.message.setText("Unknown connection error, try again later.")
                    raise e
            ei.permanent_setting("username", self.username.text())
            ei.permanent_setting("password", self.password.text())
            QtWidgets.QApplication.restoreOverrideCursor()
            return True

        def finalize(self, context: dict):
            """Store the release object in the context"""
            context["release"] = self.release

        def nextPage(self):
            """Proceed to the EcoinventVersionPage"""
            return EiWizard.EcoinventVersionPage

    class EcoinventVersionPage(widgets.ABWizardPage):
        """Wizard page to choose ecoinvent version and system model"""
        title = "Choose version"
        subtitle = "Choose ecoinvent version and system model"

        def __init__(self, parent=None):
            super().__init__(parent)

            self.versions = QtWidgets.QComboBox()
            self.models = QtWidgets.QComboBox()

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.versions)
            layout.addWidget(self.models)

            self.setLayout(layout)

        def initializePage(self, context: dict):
            """Initialize the page with available versions and models"""
            self.release = context["release"]
            self.versions.currentTextChanged.connect(self.collect_models)
            self.versions.addItems(self.release.list_versions())

        def finalize(self, context: dict):
            """Store the selected version and model in the context"""
            context["version"] = self.versions.currentText()
            context["model"] = self.models.currentText()

        def nextPage(self):
            """Proceed to the EcoinventDownloadPage"""
            return EiWizard.EcoinventDownloadPage

        def collect_models(self, version: str):
            """Collect and display system models for the selected version"""
            QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
            self.models.clear()
            self.models.addItems(self.release.list_system_models(version))
            QtWidgets.QApplication.restoreOverrideCursor()

    class EcoinventDownloadPage(widgets.ABThreadedWizardPage):
        """Wizard page to download the selected ecoinvent release"""
        title = "Download ecoinvent"
        subtitle = "Downloading the selected ecoinvent release"

        class Thread(threading.ABThread):
            """Thread to handle the download process"""
            download_ready: SignalInstance = Signal(str)

            def run_safely(self, release: ei.release, version: str, model: str):
                """Download the ecoinvent release"""
                path = release.get_release(
                    version=version,
                    system_model=model,
                    release_type=ei.ReleaseType.ecospold,
                    extract=False,
                    fix_version=False
                )
                path = str(path)
                if not path.endswith(".7z"):
                    path = path + ".7z"
                self.download_ready.emit(path)

        def __init__(self, parent=None):
            super().__init__(parent)
            self.ei_filepath = None

        def initializePage(self, context: dict):
            """Start the download thread"""
            self.thread.start(context["release"], context["version"], context["model"])
            self.thread.download_ready.connect(self.download_ready)

        def download_ready(self, filepath: str):
            """Handle the completion of the download"""
            self.ei_filepath = filepath

        def finalize(self, context: dict):
            """Store the downloaded file path in the context"""
            context["ei_filepath"] = self.ei_filepath

        def nextPage(self):
            """Proceed to the BiosphereSetupPage"""
            return EiWizard.BiosphereSetupPage

    class BiosphereSetupPage(widgets.ABWizardPage):
        """Wizard page to choose biosphere setup options"""
        title = "Biosphere setup"
        subtitle = "Choose whether to import the biosphere database or connect to an existing one"

        def __init__(self, parent=None):
            super().__init__(parent)

            self.biosphere_choice = widgets.ABRadioButtonCollapser(self)
            self.biosphere_choice.buttonClicked.connect(lambda: self.completeChanged.emit())

            self.biosphere_choice.addOption(
                name="existing",
                label="Link to an existing biosphere",
                w=widgets.ABComboBox.get_database_combobox()
            )

            self.biosphere_choice.addOption(
                name="import",
                label="Import included biosphere",
                w=widgets.DatabaseNameEdit(database_preset="biosphere")
            )

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.biosphere_choice)
            self.setLayout(layout)

        def isComplete(self):
            """Check if a biosphere option has been selected"""
            return self.biosphere_choice.currentOption() is not None

        def finalize(self, context: dict):
            """Store the selected biosphere option in the context"""
            if self.biosphere_choice.currentOption() == "existing":
                context["biosphere_name"] = self.biosphere_choice.view("existing").currentText()
            else:
                context["biosphere_name"] = self.biosphere_choice.view("import").text()

        def nextPage(self):
            """Proceed to the appropriate next page based on the biosphere choice"""
            if self.biosphere_choice.currentOption() == "existing":
                return EiWizard.EcoinventSetupPage
            else:
                return EiWizard.BiosphereInstallPage

    class BiosphereInstallPage(widgets.ABThreadedWizardPage):
        """Wizard page to install the biosphere database"""
        title = "Installing biosphere database"
        subtitle = "Installing bundled biosphere database into the project"

        class Thread(threading.ABThread):
            """Thread to handle the biosphere installation process"""
            def run_safely(self, ei_filepath: str, biosphere_name: str):
                """Install the biosphere database"""
                importer = Ecoinvent7zImporter(ei_filepath)
                importer.install_biosphere(biosphere_name)

        def initializePage(self, context: dict):
            """Start the biosphere installation thread"""
            self.thread.start(context["ei_filepath"], context["biosphere_name"])

        def nextPage(self):
            """Proceed to the MethodsSetupPage"""
            return EiWizard.MethodsSetupPage

    class MethodsSetupPage(widgets.ABWizardPage):
        """Wizard page to choose methods setup options"""
        title = "Methods setup"
        subtitle = "Choose whether to import methods from ecoinvent or from file"

        def __init__(self, parent=None):
            super().__init__(parent)

            self.methods_choice = widgets.ABRadioButtonCollapser(self)
            self.methods_choice.buttonClicked.connect(lambda: self.completeChanged.emit())

            self.methods_choice.addOption(
                name="remote",
                label="Download methods from ecoinvent",
                w=QtWidgets.QWidget()
            )

            self.methods_choice.addOption(
                name="local",
                label="Import methods from file",
                w=widgets.ABFileSelector(filter="*.xlsx")
            )

            self.methods_choice.addOption(
                name="skip",
                label="Don't import methods",
                w=QtWidgets.QWidget()
            )

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.methods_choice)

            self.setLayout(layout)

        def finalize(self, context: dict):
            """Store the selected methods option in the context"""
            if self.methods_choice.currentOption() == "remote":
                file = ei.get_excel_lcia_file_for_version(context["release"], context["version"])
                context["methods_filepath"] = str(file)
            if self.methods_choice.currentOption() == "local":
                context["methods_filepath"] = self.methods_choice.view("local").text()

        def isComplete(self):
            """Check if a methods option has been selected"""
            return self.methods_choice.currentOption() is not None

        def nextPage(self):
            """Proceed to the appropriate next page based on the methods choice"""
            if self.methods_choice.currentOption() == "remote" or self.methods_choice.currentOption() == "local":
                return EiWizard.MethodsInstallPage
            else:
                return EiWizard.EcoinventSetupPage

    class MethodsInstallPage(widgets.ABThreadedWizardPage):
        """Wizard page to install the selected methods"""
        title = "Installing methods"
        subtitle = "Installing selected methods and linking to the biosphere"

        class Thread(threading.ABThread):
            """Thread to handle the methods installation process"""
            def run_safely(self, methods_filepath: str, biosphere_name: str):
                """Install the methods and link to the biosphere"""
                importer = EcoinventLCIAImporter.setup_with_ei_excel(methods_filepath)
                importer.set_biosphere(biosphere_name)
                importer.apply_strategies()

                signals.method.blockSignals(True)
                signals.meta.blockSignals(True)

                old = bd.methods.deserialize()
                importer.write_methods(overwrite=True)

                signals.method.blockSignals(False)
                signals.meta.blockSignals(False)

                signals.meta.methods_changed.emit(deepcopy(bd.methods.data), old)

        def initializePage(self, context: dict):
            """Start the methods installation thread"""
            self.thread.start(context["methods_filepath"], context["biosphere_name"])

        def nextPage(self):
            """Proceed to the EcoinventSetupPage"""
            return EiWizard.EcoinventSetupPage

    class EcoinventSetupPage(widgets.ABWizardPage):
        """Wizard page to set up the ecoinvent database"""
        title = "Ecoinvent setup"
        subtitle = "Choose name for ecoinvent database"

        def __init__(self, parent=None):
            super().__init__(parent)
            self.database_name = widgets.DatabaseNameEdit(database_preset="ecoinvent")
            self.database_name.textChanged.connect(self.completeChanged)

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.database_name)
            self.setLayout(layout)

        def isComplete(self):
            """Check if a database name has been entered"""
            return bool(self.database_name.text())

        def finalize(self, context: dict):
            """Store the database name in the context"""
            context["database_name"] = self.database_name.text()

        def nextPage(self):
            """Proceed to the EcoinventInstallPage"""
            return EiWizard.EcoinventInstallPage

    class EcoinventInstallPage(widgets.ABThreadedWizardPage):
        """Wizard page to install the ecoinvent database"""
        title = "Installing ecoinvent"
        subtitle = "Installing ecoinvent database into the project"

        class Thread(threading.ABThread):
            """Thread to handle the ecoinvent installation process"""
            def run_safely(self, ei_filepath: str, database_name: str, biosphere_name: str):
                """Install the ecoinvent database"""
                importer = Ecoinvent7zImporter(ei_filepath)
                importer.install_ecoinvent(database_name, biosphere_name)

        def initializePage(self, context: dict):
            """Start the ecoinvent installation thread"""
            self.thread.start(
                context["ei_filepath"],
                context["database_name"],
                context["biosphere_name"]
            )

    pages = [
        RemoteOrLocalPage, LocalSelectPage, LoginPage, EcoinventVersionPage, EcoinventDownloadPage, BiosphereSetupPage,
        BiosphereInstallPage, MethodsSetupPage, MethodsInstallPage, EcoinventSetupPage, EcoinventInstallPage
    ]
